def getTransactions(self, time=None) -> TransactionList:
        #return super().getTransactions(time=time)
        self.rawTxs = self._getRawTransactions()

        for tx in self.rawTxs:
            time = datetime.strptime(tx["header"]["timestamp"],"%Y-%m-%dT%H:%M:%SZ")
            height = tx["data"]["height"]
            txHash = tx["data"]["txhash"]

            t = Transaction(dateTime=time, txHash=txHash, blockHeight=height)

            #1. Fee
            #if tx["header"]["chain_id"] == "cosmoshub-3":
            fee = tx["data"]["tx"]["auth_info"]["fee"]["amount"][0]["amount"] if "auth_info" in tx["data"]["tx"] else tx["data"]["tx"]["value"]["fee"]["amount"][0]["amount"]
            feeCurrency = tx["data"]["tx"]["auth_info"]["fee"]["amount"][0]["denom"] if "auth_info" in tx["data"]["tx"] else tx["data"]["tx"]["value"]["fee"]["amount"][0]["denom"]
            t.fee = Fee(Decimal(fee)/self.denominator, self.getSymbol(feeCurrency) )

            #2.Memo
            memo = tx["data"]["tx"]["value"]["memo"] if tx["header"]["chain_id"] == "cosmoshub-3" else tx["data"]["tx"]["body"]["memo"]
            t.memo = memo

            #only add fee once for every log
            feeAdded = False 
            for log in tx["data"]["logs"]:
                tmp = deepcopy(t)
                addTransaction = False
                events = log["events"] #todo: check --> no logs for failed txs
                delegated = False
                rewards = False
                for event in events:
                    if event["type"] == "delegate":
                        #tmp.note = "Delegated"
                        for attr in event["attributes"]:
                            print(attr["key"])
                            if attr["key"] == "amount":
                                print("FOUND")
                                tmp.note = "Delegated {}ATOM".format(Decimal(attr["value"])/self.denominator)
                        addTransaction = True
                        delegated = True
                    if event["type"] == "withdraw_rewards":
                        tmp.category = "Staking"
                        rewards = True
                    if event["type"] == "transfer":
                        attrs = event["attributes"]
                        recipientIdx = None
                        senderIdx = None
                        amountIdx = None
                        for idx,attr in enumerate(attrs):
                            if attr["key"] == "recipient":
                                recipientIdx = idx
                            if attr["key"] == "sender":
                                senderIdx = idx
                            if attr["key"] == "amount":
                                amountIdx = idx
                        
                        # We need at least a sender or recipient and an amount
                        assert(recipientIdx!=None or senderIdx!=None)
                        assert(amountIdx!=None)
                        
                        if recipientIdx is not None:
                            tmp.toAddress = attrs[recipientIdx]["value"]
                        if senderIdx is not None:
                            tmp.fromAddress = attrs[senderIdx]["value"]
                        
                        # parse amount
                        amount = attrs[amountIdx]["value"]
                        number, unit = self._parseAmount(amount)

                        # check for ibc token
                        if unit[0:4] == "ibc/":
                            unit, decimal = self.getIbcTokenSymbolFromIBCTokenHash(unit[4:])
                        else:
                            unit = self.getSymbol(unit)
                            decimal = self.denominator

                        #check for in or out transaction
                        if tmp.fromAddress == self.address:
                            tmp.posOut = Position(Decimal(number)/Decimal(decimal), unit)
                            addTransaction = True
                        elif tmp.toAddress == self.address:
                            tmp.posIn = Position(Decimal(number)/Decimal(decimal), unit)
                            addTransaction = True

                # Delete Fee for incoming txs or if it was already added for another transaction of the same tx hash
                if (tmp.posIn.amount != 0 and not(delegated) and not(rewards)) or feeAdded:
                    tmp.fee.amount = 0
                elif tmp.posIn.amount != 0 and delegated:
                    tmp.note += " (+auto claim reward)"
                    tmp.category = "Staking"
                    feeAdded = True
                else:
                    feeAdded = True

                if addTransaction:
                    self.txList.append(tmp)

        return self.txList


from abc import abstractclassmethod, abstractmethod


import abc
#import Transaction, TransactionList
from ..Transaction import TransactionList

class Network(metaclass=abc.ABCMeta):

    def __init__(self, blockExplorerUrl):
        self._blockExplorerUrl = blockExplorerUrl

    @property
    def BLOCK_EXPLORER_URL(self):
        return self._blockExplorerUrl

    @abc.abstractmethod
    def getTransactions(self, address, startTime=None, offset=None) -> TransactionList:
        return


class Coin(object):
    
    @property
    def symbol(self) -> str:
        """
        The symbol
        """
        return self.__symbol
    
    @symbol.setter
    def symbol(self, value:str):
       self.__symbol = value.upper()

    @property
    def name(self) -> str:
        """
        The name
        """
        return self.__name
    
    @name.setter
    def name(self, value):
       self.__name = value

    @property
    def id(self) -> str:
        """
        The id
        """
        return self.__id
    
    @id.setter
    def id(self, value):
       self.__id = value

    @property
    def contractAddress(self) -> str:
        """
        The contract address of the coin
        """
        return self.__contractAddress
    
    @contractAddress.setter
    def contractAddress(self, value):
       self.__contractAddress = value
    
    def __init__(self, symbol:str, name=None, id=None, contractAddress = None ) -> None:
        super().__init__()
        self.symbol = symbol
        self.name = name
        self.id = id 
        self.contractAddress = contractAddress

    def __repr__(self) -> str:
        return self.__symbol
    

    


from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Transaction import Position, Transaction, TransactionList, Fee
from pyCryptoTransactions.exchanges.coinbase.coinbaseApiWrapper import coinbase
from time import sleep
from decimal import Decimal

class CoinbaseImporter(Importer):
    def __init__(self, client_id="", client_secret="", redirect_url="") -> None:
        super().__init__()
        
        client_id= "13e48e06104ee337057efe67ba52272e7a2153c326d8c4d29fa019a735a11297"
        client_secret = "817d74bf0d54b701b894db91bcaee1eb45a70c81604aa2d9c5d03c452fa6be8b"
        redirect_url= "https://jens-brauer.de"

        #coinbase_oauth = coinbase.CoinbaseOAuth(client_id, client_secret, redirect_url)
        #tokens = coinbase_oauth.get_tokens('95424d8f51ed597ea6a0a127c1841c7ac66aaef48a9085212642d9b6583703e2',grant_type='refresh_token')
        #print(tokens)
        #raise

        #tokens = coinbase_oauth.get_tokens("30cc2537b7e4d5105733e632e6e1d9b1ec7425942af177609ba5503ef2170a1a")
        #print(tokens)
        
        self.client = self.authenticate(client_id, client_secret, redirect_url)
        coinbase.Coinbase.with_oauth('212d39c032a45af42e46a4a0a829a40dddce585f0d9766851993e152048cf5ef', 'a56af6849b50093fddcf2cc2a44d5317b5e436506b2dc18c13407f8b857e6e0a')
        self.accounts = []
        #self.getAllAccounts()
        self.rawTxs = []
        self._trades = TransactionList()

    def authenticate(self, client_id, client_secret, redirect_url):
        #coinbase_oauth = coinbase.CoinbaseOAuth(client_id, client_secret, redirect_url)
        
        #1. If no auth.json file:
        #print(coinbase_oauth.create_authorize_url(scopes))

        #a. Waiting for user input : code
        #tokens = coinbase_oauth.get_tokens(code)
        
        #b. save tokens to auth file 

        #2. if auth.json file:
        # load access and refresh token
        # try receiving something
        # if token expired, get new token from refresh_token:
        #tokens = coinbase_oauth.get_tokens(refresh_token, grant_type='refresh_token')
        # save tokens to file
        # return self.autehnticate(...) ---> call again

        # return coinbase.Coinbase.with_oauth(atoken, rtoken)

        return NotImplemented

    def getAllAccounts(self):
        self.accounts += self.__getPaginatedResult('/accounts')

    def getBalance(self) -> dict:
        balance = dict()
        for account in self.accounts:
            if Decimal(account["balance"]["amount"]) == 0:
                continue
            balance[account["balance"]["currency"]] = Decimal(account["balance"]["amount"])
        return balance

    def __getPaginatedResult(self, getUrl:str) -> list:
        elems = []
        data = self.client.get(getUrl)
        elems += data["data"]
        while data["pagination"]["next_starting_after"] != None:
            data = self.client.get(getUrl, params={'starting_after': data["pagination"]["next_starting_after"]})
            elems += data["data"]
            sleep(0.006)
        return elems
    
    def _getRawTransactionsPerAccount(self, accountId:str) -> list:
        rawTxs = self.__getPaginatedResult('/accounts/{}/transactions'.format(accountId))
        return rawTxs

    def _getRawTransactions(self) -> None:
        for account in self.accounts:
            self.rawTxs += self._getRawTransactionsPerAccount(account["id"])

    def _handleFiatDeposit(self, txRaw) -> Transaction:
        t = Transaction(dateTime=txRaw["created_at"], posIn=Position( Decimal(txRaw["amount"]["amount"]),  txRaw["amount"]["currency"]) )
        t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t

    def _handleFiatWithdrawal(self, txRaw) -> Transaction:
        #print(txRaw)
        t = Transaction(dateTime=txRaw["created_at"], posOut=Position( abs(Decimal(txRaw["amount"]["amount"])),  txRaw["amount"]["currency"]) )
        t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t

    def _handleExchangeDeposit(self, txRaw) -> Transaction:
        t = Transaction(dateTime=txRaw["created_at"], posOut=Position( abs(Decimal(txRaw["amount"]["amount"])),  txRaw["amount"]["currency"]) )
        t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t
    
    def _handleExchangeWithdrawal(self, txRaw) -> Transaction:
        """From Coinbase Pro to Coinbase """
        t = Transaction(dateTime=txRaw["created_at"], posIn=Position( abs(Decimal(txRaw["amount"]["amount"])),  txRaw["amount"]["currency"]) )
        t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t

    def _handleBuy(self, txRaw) -> Transaction:
        if txRaw["details"]["payment_method_name"].lower().find("wallet") != -1:
            #match found
            buyWithoutInternalWallet = False
        else:
            buyWithoutInternalWallet = True

        buy = self.client.get(txRaw["buy"]["resource_path"].replace('/v2',''))["data"]
        if buy["status"] != "completed":
            print("BUY: STATUS NOT COMPLETED {}".format(txRaw["sell"]["resource_path"]))
            return Transaction()
        #print(buy)
        t = Transaction(dateTime=txRaw["created_at"])
        t.orderId = txRaw["id"]
        t.posIn = Position( Decimal(buy["amount"]["amount"]),  buy["amount"]["currency"])
        t.posOut = Position( Decimal(buy["subtotal"]["amount"]), buy["subtotal"]["currency"])
        t.fee = Fee( Decimal(buy["fee"]["amount"]), buy["fee"]["currency"])
        t.tradeId = buy["transaction"]["id"]

        #we create a new deposit in case we have a buy without using the internal wallets
        if buyWithoutInternalWallet:
            t2 = Transaction(dateTime=txRaw["created_at"])
            t2.posIn = Position( Decimal(buy["total"]["amount"]), buy["total"]["currency"])
            t2.note = "Deposit for buy with payment method {}".format(txRaw["details"]["payment_method_name"])
            t2.orderId = "deposit/"+txRaw["id"]
            if self.txList.fromOrderId(t2.orderId) is None:
                self.txList.append(t2) #we can just add it here, because there won't be another transaction with the same id

        #t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t

    def _handleSell(self, txRaw) -> Transaction:
        sell = self.client.get(txRaw["sell"]["resource_path"].replace('/v2',''))["data"]
        if sell["status"] != "completed":
            print("SELL: STATUS NOT COMPLETED {}".format(txRaw["sell"]["resource_path"]))
            return Transaction()
        #print(buy)
        t = Transaction(dateTime=txRaw["created_at"])
        t.orderId = txRaw["id"]
        t.posOut = Position( Decimal(sell["amount"]["amount"]), sell["amount"]["currency"])
        t.posIn = Position( Decimal(sell["subtotal"]["amount"]), sell["subtotal"]["currency"])
        t.fee = Fee( Decimal(sell["fee"]["amount"]), sell["fee"]["currency"])
        t.tradeId = sell["transaction"]["id"]
        #t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t

    def _handleTrade(self, txRaw) -> Transaction:

        """The trade endpoint seems not to work
        trade = self.client.get(txRaw["trade"]["resource_path"].replace('/v2',''))["data"]
        if trade["status"] != "completed":
            print("TRADE: STATUS NOT COMPLETED {}".format(txRaw["sell"]["resource_path"]))
            return Transaction()
        """
        tradeId = txRaw["trade"]["id"]
        if self._trades.fromOrderId(tradeId):
            tmp,idx = self._trades.fromOrderId(tradeId)
        else:
            tmp = Transaction(dateTime=txRaw["created_at"])
            tmp.orderId = tradeId
            self._trades.append(tmp)
        
        t,idx = self._trades.fromOrderId(tradeId)
        amount =  Position( Decimal(txRaw["amount"]["amount"]), txRaw["amount"]["currency"])
        if amount.amount > 0:
            t.posIn = Position( amount.amount, amount.currency)
        elif amount.amount < 0:
            t.posOut = Position( abs(amount.amount), amount.currency)
        
        if "fee" in txRaw:
            t.fee = Fee( Decimal(txRaw["fee"]["amount"]), txRaw["fee"]["currency"])
        
        #self._trades.append(t)
        
        #t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t
    
    def _handleSend(self, txRaw) -> Transaction:
        """ E.g. Coinbase Earn """
        if txRaw["status"] != "completed":
            print("SEND: STATUS NOT COMPLETED {}".format(txRaw["sell"]["resource_path"]))
            return Transaction()

        t = Transaction(dateTime=txRaw["created_at"])
        amount = Decimal( txRaw["amount"]["amount"] )
        currency = txRaw["amount"]["currency"] 
        if amount > 0:
            t.posIn = Position( amount, currency)
        else:
            t.posOut = Position( abs(amount), currency)
        
        if "fee" in txRaw:
            t.fee = Fee( Decimal(txRaw["fee"]["amount"]), txRaw["fee"]["currency"])
        
        if "to" in txRaw:
            if "address" in txRaw["to"] and amount <0:
                t.toAddress = txRaw["to"]["address"]
        
        # Overwrite with network values if availabe and if they are different
        if "network" in txRaw:
            if "hash" in txRaw["network"]:
                t.txHash = txRaw["network"]["hash"]
            if "transaction_amount" in txRaw["network"]:
                amount2 = Decimal( txRaw["network"]["transaction_amount"]["amount"] )
                currency2 = txRaw["network"]["transaction_amount"]["currency"]
                if amount2 != abs(amount) and amount2 != 0:
                    if amount > 0: #amount2 is always positive
                        t.posIn = Position( amount2, currency2)
                    else:
                        t.posOut = Position( abs(amount2), currency2)
            if "transaction_fee" in txRaw["network"] and Decimal( txRaw["network"]["transaction_fee"]["amount"] ) != t.fee.amount:
                t.fee = Fee( Decimal( txRaw["network"]["transaction_fee"]["amount"] ), txRaw["network"]["transaction_fee"]["currency"] )

        t.orderId = txRaw["id"]
        t.price = Position( abs( Decimal( txRaw["native_amount"]["amount"] ) ), txRaw["native_amount"]["currency"] )
        t.note += txRaw["details"]["title"] + '. ' + txRaw["details"]["subtitle"]
        return t


    def _handleRawTx(self, txElem) -> Transaction:
        tx = Transaction()
        if txElem["type"] == "exchange_deposit":
            tx = self._handleExchangeDeposit(txElem)

        elif txElem["type"] == "exchange_withdrawal" or txElem["type"] == "pro_withdrawal":
            tx = self._handleExchangeWithdrawal(txElem)

        elif txElem["type"] == "fiat_deposit":
            tx = self._handleFiatDeposit(txElem)

        elif txElem["type"] == "fiat_withdrawal":
            tx = self._handleFiatWithdrawal(txElem)

        elif txElem["type"] == "buy":
            tx =  self._handleBuy(txElem)

        elif txElem["type"] == "sell":
            tx = self._handleSell(txElem)
        
        elif txElem["type"] == "send":
            tx = self._handleSend(txElem)

        elif txElem["type"] == "trade": #not documented in API
            tx = self._handleTrade(txElem)
        
        # Todo: Implement:
        elif txElem["type"] == "transfer":
            print("transfer")
            pass 

        elif txElem["type"] == "request":
            print("request")
            pass 
        elif txElem["type"] == "vault_withdrawal":
            print("vault")
            pass
        elif txElem["type"] == "advanced_trade_fill":
            print("advanced")
            pass

        else:
            print("UNKNOWN:")
            print(txElem)
        
        return tx
    
    def getTransactions(self, time=None) -> TransactionList:
        self.getAllAccounts()
        print("Fetched all accounts")
        sleep(0.006)
        print("Getting all txs")
        self._getRawTransactions()
        print("Done")
        sleep(0.006)

        for tx in self.rawTxs:
            if self.txList.fromOrderId(tx["id"]) is not None:
                continue
            if tx["type"] == "trade":
                #trades are a bit buggy, so we handle them in a different list
                self._handleRawTx(tx)
            else:
                tmp = self._handleRawTx(tx)
                if tmp.tradeId != '' and self.txList.fromTradeId(tmp.tradeId) is not None:
                    continue
                else:
                    self.txList.append( self._handleRawTx(tx) )

        self.txList += self._trades
        
        self.txList.sort(key=lambda r: r.datetime)

        return self.txList



        #TODO: check payment method ---> credit card should create a new deposit entry


from decimal import Decimal
from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer

import iota #pip install pyota
import datetime
import pytz as timezone

class IotaImport(Importer):

    def __init__(self, address):
        super().__init__()
        self.address = iota.Address(address[0:81].upper())
        self.api = iota.Iota('https://nodes.iota.org:443')
        self.rawTxs = []

    def _findTransactions(self):
        response = self.api.find_transaction_objects(addresses=[self.address])
        if not response['transactions']:
            print('Couldn\'t find data for the given address.')
        else:
            self.rawTxs = response["transactions"]

    def getTransactions(self, startTime=None, offset=None) -> TransactionList:
        self._findTransactions()
        for tx in self.rawTxs:
            if tx.value != 0:
                time = datetime.datetime.fromtimestamp(tx.timestamp, tz=timezone.utc)
                t = Transaction(dateTime=time)
                if tx.value > 0:
                    t.posOut = Position(Decimal(tx.value)/Decimal(1E6), "IOTA")
                else:
                    t.posIn = Position(Decimal(tx.value)/Decimal(1E6), "IOTA")
                t.txHash = tx.hash.as_json_compatible()
                self.txList.append(t)
        
        return self.txList


from .CoinbaseAPIKeyAuthentication import CoinbaseAPIKeyAuthentication
from .CoinbaseOAuth import CoinbaseOAuth
from .CoinbaseOAuthAuthentication import CoinbaseOAuthAuthentication
from .CoinbaseRPC import CoinbaseRPC
from .error import CoinbaseAPIException
from .error import CoinbaseException
class Coinbase(object):
    '''
    Manages all data and interactions with the Coinbase API.
    '''

    @staticmethod
    def with_api_key(key, secret, nonce=None):
        return Coinbase(
                CoinbaseAPIKeyAuthentication(
                    key, secret),
                nonce)

    @staticmethod
    def with_oauth(access_token, refresh_token):
        return Coinbase(
                CoinbaseOAuthAuthentication(
                    access_token, refresh_token))

    def __init__(self, authentication, nonce=None):
        self.__authentication = authentication
        self.__rpc = CoinbaseRPC(self.__authentication, nonce)

    def delete(self, path, params=None):
        return self.__rpc.request('DELETE', path, params)

    def get(self, path, params=None):
        return self.__rpc.request('GET', path, params)

    def post(self, path, params=None):
        return self.__rpc.request('POST', path, params)

    def put(self, path, params=None):
        return self.__rpc.request('PUT', path, params)


from .error import CoinbaseAPIException

import json
import requests
import urllib.parse
class CoinbaseOAuth(object):
    '''
    Handles client OAuth functionality like authorizing a user's account to be
    accessed.
    '''

    def __init__(self, client_id, client_secret, redirect_uri):
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__redirect_uri = redirect_uri


    def create_authorize_url(self, scope=list()):
        params = {
            'response_type': 'code',
            'client_id': self.__client_id,
            #'client_secret': self.__client_secret,
            'redirect_uri': self.__redirect_uri,
            'account': 'all' #grants access to all accounts and not just one, see https://developers.coinbase.com/docs/wallet/coinbase-connect/reference
        }

        for i in range(len(scope)):
            scope[i] = urllib.parse.quote_plus(scope[i])

        params = urllib.parse.urlencode(params) + '&scope=' + '+'.join(scope)

        return 'https://coinbase.com/oauth/authorize?' + params


    def get_tokens(self, code, grant_type='authorization_code'):
        params = {
            'grant_type': grant_type,
            'redirect_uri': self.__redirect_uri,
            'client_id': self.__client_id,
            'client_secret': self.__client_secret
        }

        if grant_type == 'refresh_token':
            params['refresh_token'] = code
        else:
            params['code'] = code

        request = \
            requests.post(
                'https://coinbase.com/oauth/token',
                data=params)

        if request.status_code != 200:
            raise CoinbaseAPIException(
                    'Could not get tokens - code',
                    request.status_code,
                    request.content.decode())

        return json.loads(request.content.decode())


    def refresh_tokens(self, refresh_token):
        return self.get_tokens(refresh_token, 'refresh_token')



#!/usr/bin/env python3

from math import log, ceil
import sys

#Info:https://bitcoin.stackexchange.com/questions/28380/i-want-to-generate-a-bip32-version-number-for-namecoin-and-other-altcoins

# lookup tables to convert integers in the range [0, 58) to base-58 digits and back
int_to_b58_digit = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
b58_digit_to_int = { b58:i for i,b58 in enumerate(int_to_b58_digit) }

# convert a (long) integer to its base-58 representation string
def int_to_base58_str(int_rep):
    base58_str = ''
    while int_rep:
        int_rep, remainder = divmod(int_rep, 58)
        base58_str = int_to_b58_digit[remainder] + base58_str
    return base58_str


def prepended_bytes(prepended_b58_digits, b256_digit_count):

    # ones are a special case in base58check format;
    # count and remove them, they are added back in later
    ones = 0
    for b58 in prepended_b58_digits:
        if b58 != '1':
            break
        ones += 1
        prepended_b58_digits = prepended_b58_digits[1:]
    if not prepended_b58_digits:  # if they're all 1's
        return ones * b'\0'

    # calc the # of base58 digits required for b256_digit_count bytes of "real" data
    # (not including the prepended base58 digits)
    b58_digit_count = ceil(b256_digit_count * log(256) / log(58))

    do_overflow_check = True
    while True:
        # calc the minimum integral value which starts with the desired digits in base-58
        min_int = 0
        for b58 in prepended_b58_digits:
            min_int *= 58
            min_int += b58_digit_to_int[b58]
        #
        # left-shift (mult. by a power of 58) to be just left of the "real" base-58 data
        min_int *= 58 ** b58_digit_count

        # uncomment to sanity-check that min_int is correct
        #print(" min_int:", ones * '1' + int_to_base58_str(min_int))

        # right-shift by multiples of 8 bits (base-256) to retrieve only the
        # most-significant bytes which are to the left of the "real" base-256 data
        min_int >>= b256_digit_count * 8
        # right-shifing effectively rounds min_int down, but we
        # need it rounded up, so add one to round it up instead
        min_int += 1

        # because min_int has been rounded up above, it's possible that adding it to "real"
        # data could cause an overflow in base-58 making the prepended_b58_digits increment
        # by one; if this could happen, left-shift prepended_b58_digits and repeat
        if do_overflow_check:
            max_real_data_int = (1 << b256_digit_count*8) - 1
            max_base58_str = int_to_base58_str((min_int << b256_digit_count*8) + max_real_data_int)
            if not max_base58_str.startswith(prepended_b58_digits):
                prepended_b58_digits += '1'
                do_overflow_check = False  # it doesn't matter if the '1' appended above overflows to '2'

                # uncomment to confirm that the max possible value
                # wouldn't have the desired prepended base-58 digits
                #print("overflow:", ones * '1' + max_base58_str)

                continue

        # prepend any ones according to base58check rules, and convert min_int to a byte string
        return ones * b'\0' + min_int.to_bytes(length= (min_int.bit_length() + 7) // 8, byteorder='big')


if __name__ == '__main__':
    if len(sys.argv) != 3:
        sys.exit('usage: {} <STRING-TO-PREPEND> <DATA-BYTE-LEN (excluding 4-byte checksum)>'.format(sys.argv[0]))

    result = prepended_bytes(sys.argv[1], int(sys.argv[2]) + 4)  # add 4 for the checksum
    print(', '.join('{:#04x}'.format(i) for i in result))


from etherscan.accounts import Account
from etherscan.client import EmptyResponse
import datetime
from django.utils import timezone
from ..portfolioConfig import ETHERSCAN_APIKEY

class EtherTransactions():
    def __init__(self,  etherAccountAddress, apiKey=ETHERSCAN_APIKEY):
        self.apiKey = apiKey
        self.address = etherAccountAddress.lower()
        self.etherScanAccountApi = Account(address=etherAccountAddress, api_key=self.apiKey)
        self.transactions = []
        self.erc20Transactions = []
        self.internalTransactions = []

    def update(self):
        try:
            self.transactions = self.etherScanAccountApi.get_transaction_page()
        except EmptyResponse:
            self.transactions = []
        try:
            self.erc20Transactions = self.etherScanAccountApi.get_transaction_page(erc20=True)
        except EmptyResponse:
            self.erc20Transactions = []
        try:
            self.internalTransactions = self.etherScanAccountApi.get_transaction_page(erc20=False, internal=True)
        except EmptyResponse:
            self.internalTransactions = []

    def getAllTransactions(self, startTime=None):
        self.update()
        normalTransactions = self.getTransactionsAsDict(erc20=False)
        erc20TransActions = self.getTransactionsAsDict(erc20=True)
        internalTransactions = self.getTransactionsAsDict(erc20=False, internal=True)
        allTransactions = normalTransactions + erc20TransActions + internalTransactions

        # Check for swap:
        for normalTransaction in normalTransactions:
            for erc20TransAction in erc20TransActions:
                if normalTransaction["txId"] == erc20TransAction["txId"] and normalTransaction["time"] == erc20TransAction["time"]:
                    if erc20TransAction["category"] == "Trade":
                        raise("Unknown condition")
                    if erc20TransAction["category"] == "Transfer In":
                        erc20TransAction["category"] = "Swap"
                        erc20TransAction["fee"] = normalTransaction["fee"]
                        erc20TransAction["feeCurrency"] = normalTransaction["feeCurrency"]
                        erc20TransAction["amountOut"] = normalTransaction["amountOut"]
                        erc20TransAction["currencyOut"] = normalTransaction["currencyOut"]
                        allTransactions.remove(normalTransaction)
                    elif erc20TransAction["category"] == "Transfer Out":
                        erc20TransAction["category"] = "Swap"
                        erc20TransAction["amountIn"] = normalTransaction["amountIn"]
                        erc20TransAction["currencyOut"] = normalTransaction["currencyIn"]
                        allTransactions.remove(normalTransaction)

        # Remove elements that are before startTime
        # for transaction in allTransactions:

        return allTransactions

    def isTransactionInERC20Transactions(self, transaction):
        for erc20Transaction in self.erc20Transactions:
            if erc20Transaction["hash"] == transaction["hash"]:
                return True
        return False

    def getTransactions(self,erc20=False, internal=False):
        if erc20:
            return self.erc20Transactions
        elif internal:
            return self.internalTransactions
        else:
            return self.transactions

    def getTransactionsAsDict(self, erc20=False, internal=False):
        transactionList = []
        for transaction in self.getTransactions(erc20, internal):
            entry = {}
            value = int(transaction["value"])/1E18 if not erc20 else int(transaction["value"]) / (10**int(transaction["tokenDecimal"]))
            currency = "ETH" if not erc20 else transaction["tokenSymbol"]

            if transaction["from"] == self.address and value > 0:
                entry["category"] = "Transfer Out"
                entry["amountOut"] = value
                entry["currencyOut"] = currency
                # todo check for erc20 token with same txid, then its a swap! ###or: search in both output transactions fpr same id!!

            elif transaction["to"] == self.address and value > 0:
                entry["category"] = "Transfer In"
                entry["amountIn"] = value
                entry["currencyIn"] = currency
                # todo check for erc20 token with same txid, then its a swap!

            elif transaction["from"] == self.address and value == 0:
                # Check if an complimentary entry exists in the ERC20 transactions
                # Then skip, because fees are contained in that
                if not(erc20) and self.isTransactionInERC20Transactions(transaction):
                    continue
                entry["category"] = "Fees"
            else:
                print(transaction)
                print(value)
                print(transaction["to"] == self.address)
                raise ("Unknown transaction!")

            entry["txId"] = transaction["hash"]
            entry["time"] = datetime.datetime.fromtimestamp(int(transaction["timeStamp"]), tz=timezone.utc)

            if not(erc20) and bool(int(transaction["isError"])):
                entry["category"] = "Loss"

            if erc20 and len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"] and \
                    transactionList[-1]["category"] != entry["category"]:
                lastEntry = transactionList.pop()
                entry = {**lastEntry, **entry}
                entry["category"] = "Trade"

            if entry["category"] != "Transfer In":
                entry["fee"] = int(transaction["gasUsed"]) * int(transaction["gasPrice"]) / 1E18
                entry["feeCurrency"] = "ETH"

            if internal:
                entry["note"] = "Internal Tx"

            transactionList.append(entry)

        return transactionList



from datetime import date
from decimal import Decimal
import pandas as pd

from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Transaction import Position, Transaction, TransactionList, Fee

class CSVImporterDefault(Importer):
    
    def __init__(self, filename):
        super().__init__()
        self.filename = filename
        self.df =  pd.read_csv(self.filename)
    
    def parseCSV(self):
        txs = TransactionList()
        #txs.fillFromDataframe(self.df)
        return txs
    



# ----- Author ----------------------------------------------------------------

__author__ = 'Michael Montero <mike@resy.com>'

# ----- Public Classes --------------------------------------------------------

class CoinbaseException(Exception):

    def __init__(self, message, http_code=None, response=None):
        self.message = message
        self.http_code = http_code
        self.response = response


    def get_http_code(self):
        return self.http_code


    def get_message(self):
        return self.message


    def get_response(self):
        return self.response


class CoinbaseAPIException(CoinbaseException):
    pass



import abc
import pandas as pd
import datetime
from decimal import Decimal
import numpy as np
import json
from typing import Union
from .Coin import Coin

class Position(object):
    """ A position represents an amount and its currency. """

    def __init__(self, amount, currency):
        self.amount = amount
        self.currency = currency

    @property
    def amount(self) -> Decimal:
        """
        The position amount
        Returns:
            float: The position amount
        """
        return self.__amount
    
    @amount.setter
    def amount(self, amount:str | Decimal | float | int):
        if not(isinstance(amount, Decimal)):
            if isinstance(amount,str):
                amount = Decimal(amount)
            else:
                amount = Decimal(str(amount))
        if amount.is_nan():
            return
        else:
            self.__amount = amount

    @property
    def currency(self) -> str:
        """
        The position currency.
        Returns:
            str: The name of the currency.
        """
        return self.__currency
    
    @currency.setter
    def currency(self, value:str|Coin):
        if isinstance(value, Coin):
            self.__currency = value
            return
        if isinstance(value, str):
            self.__currency = Coin(value.upper())
            return 
        if not(isinstance(value, str)):
            value = Coin(str(value).upper())
            self.__currency = value
            return 
        if value == "nan":
            return
        else:
            BaseException("Coin: Wrong setting of Coin")
    
    def __add__(self, new):
        """
        Add two positions.
        Returns:
            Position object
        """
        if isinstance(new,Decimal) and new.is_nan():
            return self
        # Don't add two currencies if we already have a balance != 0
        if self.amount != 0:
            assert(new.currency == self.currency)
        amount = self.amount + new.amount
        return Position(amount, new.currency)
        #self.amount += new.amount
        #print("added:")
        #print(self.amount)
    
    def __eq__(self, __o: object) -> bool:
        """Overrides the default implementation"""
        if isinstance(__o, Position):
            return self.amount == __o.amount and self.currency.symbol == __o.currency.symbol
        return NotImplemented
    
    def __repr__(self):
        return ", ".join([str(self.amount),self.currency])


class Fee(Position):
    """ The fee of a transaction """
    pass


class Transaction(object):
    """
    Transaction Object
    """

    @property
    def datetime(self):
        """
        The date and time of the transaction
        Returns:
            datetime.datetime: The datetime of the transaction
        """
        return self.__datetime
    
    @datetime.setter
    def datetime(self, dateTime):
        if isinstance(dateTime, datetime.datetime):
            self.__datetime = dateTime
        elif isinstance(dateTime, str):
            if dateTime[-1] == "Z":
                dateTime = dateTime[:-1]
            try:
                self.__datetime = datetime.datetime.fromisoformat(dateTime)
            except ValueError:
                print("Iso Format not detected")
        
    @property
    def posIn(self):
        return self.__posIn
    
    @posIn.setter
    def posIn(self, posIn):
        self.__posIn = posIn

    @property
    def posOut(self):
        return self.__posOut
    
    @posOut.setter
    def posOut(self, posOut):
        self.__posOut = posOut

    @property
    def txHash(self):
        return self.__txHash
    
    @txHash.setter
    def txHash(self, hash):
        self.__txHash = hash

    @property
    def tradingPair(self):
        """
        A pair of currencies representing the trading direction
        Returns:
            tuple[Position, Position]: A tuple with two positions
        """
        return self.__tradingPair
    
    @tradingPair.setter
    def tradingPair(self, pair):
        assert len(pair) == 2, "No pair provided"
        self.__tradingPair = pair
    
    @property
    def price(self):
        """
        The price per coin in fiat currency.
        Returns:
            float: The price of a coin.
        """
        return self.__price
    
    @price.setter
    def price(self, value):
        self.__price = value

    @property
    def fee(self) -> Fee:
        """
        The trading fee
        Returns:
            Fee: The trading fee
        """
        return self.__fee
    
    @fee.setter
    def fee(self, feeValue):
        assert isinstance(feeValue, Fee), 'The fee has to be of type Fee.'
        self.__fee = feeValue

    @property
    def category(self) -> str:
        """
        The category
        """
        return self.__category
    
    @category.setter
    def category(self, value):
       self.__category = value

    @property
    def note(self) -> str:
        return self.__note

    @note.setter
    def note(self, value:str):
        self.__note = value

    @property
    def orderId(self) -> str:
        return self.__orderId

    @orderId.setter
    def orderId(self, value:str):
        self.__orderId = value
    
    def isMultiSend(self) -> bool:
        return self._isMultiSend
    
    def isOutBound(self) -> bool:
        return self.posOut is not None and self.posOut.amount > 0

    def __init__(self, dateTime = None,
                       price:Decimal = Decimal(0),
                       posIn:Position = Position(0,"EUR"),
                       posOut:Position = Position(0,"EUR"),
                       fee:Fee = Fee(0,"EUR"),
                       txHash:str = "",
                       category:str = "",
                       blockHeight:int = 0,
                       tradingPair:tuple = [None, None],
                       network:str = "",
                       wallet:str = ""):
        '''
        The __init__ function is called when you create a new object
        
        :param dateTime: The time of the trade
        :param price: The price of the trade
        :type price: Decimal
        :param posIn: The position you bought the tokens
        :type posIn: Position
        :param posOut: The amount of the currency that you received
        :type posOut: Position
        :param fee: The fee that was paid for the trade
        :type fee: Fee
        :param txHash: The hash of the transaction that this trade is part of
        :type txHash: str
        :param category: The category of the transaction. Can be one of the following:
        :type category: str
        :param blockHeight: The block number of the block the transaction was included in, defaults to 0
        :type blockHeight: int (optional)
        :param tradingPair: A tuple of the form (base, quote) where base and quote are the names of the assets
        :type tradingPair: tuple
        :param network: The network that the transaction was made on
        :type network: str
        :param wallet: The wallet that initiated the transaction
        :type wallet: str
        '''

        super().__init__()

        if dateTime == None:
            self.__datetime = datetime.datetime.now()
        else:
            self.datetime = dateTime
        self.__price = price
        self.posIn = posIn
        self.posOut = posOut
        self.txHash = txHash
        self.category = category
        self.blockHeight = blockHeight
        self.tradingPair = tradingPair
        self.network = network
        self.wallet = wallet
        self.note = ""

        self._isMultiSend = False

        #exchange info
        self.exchange = ""
        self.orderId = ""
        self.tradeId = ""
        self.memo = ""

        #DEX
        self.contractAddress = ""

        #Others
        self.fromAddress = ""
        self.toAddress = ""

        if fee is None:
            self.fee = Fee(0, "EUR")
        else:
            assert isinstance(fee, Fee), 'The fee has to be of type Fee.'
            self.fee = fee

    def __repr__(self):
        return ", ".join([self.datetime.strftime("%Y/%m/%d, %H:%M:%S"),
                          str(self.posIn.amount) if self.posIn else "NONE",
                          self.posIn.currency.symbol,
                          str(self.posOut.amount) if self.posOut else "NONE",
                          self.posOut.currency.symbol,
                          self.txHash,
                          #self.trading_pair[0],
                          #self.trading_pair[1],
                          #str(self.price),
                          str(self.fee.amount),
                          self.fee.currency.symbol, self.note])
    
    def asDict(self):
        d = dict()
        d["Date"] = self.datetime 
        d["Sent Amount"] = self.posOut.amount if self.posOut.amount != 0 else None
        d["Sent Currency"] = self.posOut.currency if self.posOut.amount != 0 else ""
        d["Received Amount"] = self.posIn.amount if self.posIn.amount != 0 else None
        d["Received Currency"] = self.posIn.currency if self.posIn.amount != 0 else ""
        d["Fee Amount"] = self.fee.amount if self.fee.amount != 0 else None
        d["Fee Currency"] = self.fee.currency if self.fee.amount != 0 else ""
        d["TxHash"] = self.txHash
        d["Wallet"] = self.wallet
        d["Network"] = self.network
        d["Block Height"] = self.blockHeight
        d["Memo"] = self.memo
        d["Label"] = self.category
        d["Description"] = self.note
        d["fromAddress"] = self.fromAddress
        d["toaddress"] = self.toAddress
        d["exchange"] = self.exchange
        d["orderId"] = self.orderId

        return d

    def combinePositions(self, t):
        if not(isinstance(t, Transaction)):
            return NotImplemented
        if self.posIn.amount == 0 and t.posIn.amount != 0:
            self.posIn = t.posIn 
        if self.posOut.amount == 0 and t.posOut.amount != 0:
            self.posOut = t.posOut
        # todo; check other fields as well
    
    def serialize(self, obj):
        """JSON serializer for objects not serializable by default json code"""
        if isinstance(obj, datetime.datetime):
            serial = obj.isoformat()
            return serial

        if isinstance(obj, Decimal):
            serial = str(obj)
            return serial

        return obj.__dict__
    
    def toJSON(self):
        return json.dumps(self, default=lambda o: self.serialize(o), 
            sort_keys=True, indent=4)

class TransactionList(list):
    def __init__(self):

        super().__init__()

    def hasTx(self, txHash):
        for transaction in self:
            if transaction.txHash == txHash:
                return True
        return False
    
    def fromTxHash(self, txHash):
        for idx,transaction in enumerate(self):
            if transaction.txHash == txHash:
                return transaction,idx
        return None
    
    def fromOrderId(self, orderId):
        for idx,transaction in enumerate(self):
            if transaction.orderId == orderId:
                return transaction,idx
        return None
    
    def fromTradeId(self, tradeId):
        for idx,transaction in enumerate(self):
            if transaction.tradeId == tradeId:
                return transaction,idx
        return None
    
    def toPandasDataframe(self):
        data = []

        for elem in self:
            if elem is not None:
                data.append(elem.asDict())

        df = pd.DataFrame(data)

        return df
    
    def calculateBalance(self):
        balance = dict()
        for asset in self.getAssetList():
            amount = Decimal(0)
            for tx in self:
                amount += tx.posIn.amount if tx.posIn.currency == asset else 0
                amount -= tx.posOut.amount if tx.posOut.currency == asset else 0
                amount -= tx.fee.amount if tx.fee.currency == asset else 0
            if amount != 0 and not(amount.is_nan()):
                balance[asset] = amount
        return balance

    def getAssetList(self):
        assets = []
        for tx in self:
            if tx.posIn.currency.symbol != "":
                assets.append(tx.posIn.currency.symbol)
            if tx.posOut.currency.symbol != "":
                assets.append(tx.posOut.currency.symbol)
            if tx.fee.currency.symbol != "":
                assets.append(tx.fee.currency.symbol)
        assets = list(set(assets)) # unique strings
        return assets

    def getLedger(self, coin):
        for tx in self:
            if tx.posIn.currency == coin and tx.posIn.amount > 0:
                print("{}:      {}".format(tx.datetime, tx.posIn.amount))
            if tx.posOut.currency == coin and tx.posOut.amount > 0:
                print("{}:     {}".format(tx.datetime, -tx.posOut.amount))
            if tx.fee.currency == coin and tx.fee.amount > 0:
                print("{}:     {}".format(tx.datetime, -tx.fee.amount))
    
    def toFile(self, filename):
        out = "[" + ','.join(item.toJSON() for item in self) + ']'
        with open(filename, "w") as outfile:
            outfile.write(out)
        



    


from . import Importer
from .Transaction import TransactionList

class Wallet(object):

    def __init__(self, walletName) -> None:
        self.walletName = walletName
        self.importers = list[Importer.Importer]()

    def addNetworkImporter(self, importer:Importer) -> None:
        self.importers.append(importer)

    def getTxs(self) -> TransactionList:
        txs = TransactionList()
        for importer in self.importers:
            txs += importer.getTransactions()
        return txs


from requests import Request, Session
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import json
from datetime import datetime

import pytz as timezone
from decimal import Decimal
from copy import deepcopy

from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Coin import Coin
import time

class CosmosChain(Importer):
    def __init__(self, address):
        super().__init__()
        self._apiAdress = 'https://api.cosmostation.io/v1/'
        self._initRequest()
        self.rawTxs = None
        self.address = address
        self.ibcTokens = None
        self.denominator = Decimal(int(1E6))
        self.chain = "cosmoshub-4"
        self.ibcTokenApiAddress = "https://api-utility.cosmostation.io/v1//ibc/tokens/cosmoshub-4"
        self.unit = "ATOM"

        self._getAllIBCTokens()
    
    @property
    def pathToMemo(self):
        pathOld = ["data","tx","value","memo"]
        pathNew = ["data","tx","body","memo"]
        return pathOld if self.chain == "cosmoshub-3" else pathNew
    
    @property
    def pathToFee(self):
        pathOld = ["data","tx","value","fee","amount",0,"amount"]
        pathNew = ["data","tx","auth_info","fee","amount",0,"amount"]
        return pathOld if self.chain == "cosmoshub-3" else pathNew
    
    @property
    def pathToFeeCurrency(self):
        pathOld = ["data","tx","value","fee","amount",0,"denom"]
        pathNew = ["data","tx","auth_info","fee","amount",0,"denom"]
        return pathOld if self.chain == "cosmoshub-3" else pathNew
    
    # Endpoint doesn't give multisend transactions!
    def _getRawTransactions(self, offset=0, limit=50, **kwargs):#offset=0, limit=500, startTime=None, endTime = int(round(time.time() * 1000))):
        values = {}
        values['limit'] = limit
        values['from'] = offset
        for k in kwargs:
            if kwargs[k] is not None:
                values[k] = kwargs[k]
        return self._request('/account/new_txs/{}'.format(self.address), **values)
    
    def _getAllIBCTokens(self):
        data= self._request('',apiAdress=self.ibcTokenApiAddress)
        self.ibcTokens = data
        print("Fetched {} IBC Tokens".format(len(self.ibcTokens["ibc_tokens"])))
    
    def getIbcTokenSymbolFromIBCTokenHash(self, ibcTokenHash):
        symbol = "ibc/"+ibcTokenHash
        decimal = self.denominator
        for token in self.ibcTokens["ibc_tokens"]:
            if token["hash"] == ibcTokenHash.strip():
                if "decimal" in token:
                    decimal = '1' + int(token["decimal"])*'0'
                if "display_denom" in token:
                    symbol = token["display_denom"]
                elif "base_denom" in token:
                    symbol = token["base_denom"]
                return symbol, decimal
        print("Could not find token name for IBC token {}".format(ibcTokenHash))
        return symbol,decimal
    
    def getSymbol(self, symbol):
        if symbol == "uatom":
            return "ATOM"
        if symbol == "uosmo":
            return "OSMO"
        if symbol == "uion":
            return "ION"
        if symbol == "uxprt":
            return "XPRT"
        return symbol
    
    def _getAllRawTransactions(self):
        offset = 0
        rawData = self._getRawTransactions(offset=offset, limit=50)
        numTxs = len(rawData)
        while numTxs == 50:
            print("Fetching another page starting from block id {}".format(rawData[-1]["header"]["block_id"]))
            lastBlock = rawData[-1]["header"]["block_id"]
            tmp = self._getRawTransactions(offset=lastBlock, limit=50)
            numTxs = len(tmp)
            if tmp:
                rawData.extend(tmp) #todo:check
        return rawData
    
    def _parseAmount(self, amountStr) -> str:
        """Parse amount values like 1234uatom"""
        if amountStr.isdigit():
            #no unit
            number = amountStr
            unit = self.unit 
        else:
            for i,c in enumerate(amountStr):
                if not c.isdigit():
                    break
            number = amountStr[:i]
            unit = amountStr[i:].lstrip()
        return number, unit

    def _parseAmounts(self, amountStr) -> list:
        amounts = amountStr.split(',')
        amountList = list()
        for amount in amounts:
            number, unit = self._parseAmount(amount)
            contractAddress = None
            symbol = unit
            if unit == '' and float(number) > 0:
                symbol = self.unit
            # check for ibc token
            if unit[0:4] == "ibc/":
                symbol, decimal = self.getIbcTokenSymbolFromIBCTokenHash(unit[4:])
                contractAddress = unit
            else:
                symbol = self.getSymbol(unit)
                decimal = self.denominator
            
            coin = Coin(symbol=symbol, contractAddress=contractAddress)
            
            amountList.append(Position(Decimal(number)/Decimal(decimal), coin))
        return amountList

    def _findEventByType(self, allEvents, type:str):
        for event in allEvents:
            if event["type"] == type:
                return event
        return None
    
    def handleEvents(self, events, currentTransactionObject):
        txList = self.handleTransferEvent(events, currentTransactionObject)
        isDelegated, delegatedAmount = self.handleDelegationEvent(events)
        hasReward, rewards = self.handleRewardEvent(events)
        hasProposal = self.handlePropsoalEvent(events)
        hasClaim, claims = self.handleClaimEvent(events) #more for osmosis then for cosmos

        # If we have a delegation event and a tx input/output(?) then we have a auto claim reward
        if txList and isDelegated:
            for tx in txList:
                tx.note += "Delegated {}{}".format(delegatedAmount.amount, delegatedAmount.currency) + " (+auto claim reward)"
                tx.category = "Staking"

        # If we have no input and ouput txs but a delegated event, we only have transaction fees -> we create a new txlist with one entry
        if not txList and isDelegated:
            txList = TransactionList()
            currentTransactionObject.note = "Delegated {}{}".format(delegatedAmount.amount, delegatedAmount.currency)
            txList.append(currentTransactionObject)
        
        # If Rewards are included, find the transaction with the reward and mark it as "Staking"
        if hasReward:
            for tx in txList:
                for reward in rewards:
                    if tx.posIn == reward:
                        tx.category = "Staking"
                        tx.note += "Staking Reward"
        
        if hasClaim:
            for tx in txList:
                for claim in claims:
                    if tx.posIn == claim:
                        tx.category = "Staking"
                        tx.note += "Airdrop (?)"


        #Remove fees for received txs only:
        for tx in txList:
            if (tx.posIn.amount != 0 and not(isDelegated) and not(hasReward) and not (hasProposal) and not(hasClaim)): #when it has a reward we must have sent a tx fee
                tx.fee.amount = 0
        
        return txList

    def handleDelegationEvent(self, events):
        isDelegated = False
        delegatedAmount = Position(0,"")
        event = self._findEventByType(events,"delegate")
        if event != None:
            isDelegated = True 
            for attr in event["attributes"]:
                if attr["key"] == "amount":
                    delegatedAmounts = self._parseAmounts(attr["value"])
                    assert(len(delegatedAmounts) == 1) #there should only be one delegation at a time
                    delegatedAmount = delegatedAmounts[0]
        return isDelegated, delegatedAmount
    
    def handleRewardEvent(self, events):
        event = self._findEventByType(events,"withdraw_rewards")
        rewards = list()
        if event:
            for attr in event["attributes"]:
                if attr["key"] == "amount":
                    rewards = self._parseAmounts(attr["value"])
            return True, rewards
        else:
            return False, rewards

    def handlePropsoalEvent(self, events):
        event = self._findEventByType(events, "proposal_vote")
        if event:
            return True 
        else:
            return False
    
    def handleClaimEvent(self, events):
        event = self._findEventByType(events, "claim")
        claims = list()
        if event:
            for attr in event["attributes"]:
                if attr["key"] == "amount":
                    claims = self._parseAmounts(attr["value"])
            return True, claims
        else:
            return False, claims

    def handleTransferEvent(self, events, currentTransactionObject) -> TransactionList:
        event = self._findEventByType(events,"transfer")
        if event == None:
            return TransactionList()

        transferAttributes = event["attributes"]
        numAttributes = len(transferAttributes)
        txList = TransactionList()
        start = 0
        while(start+1 < numAttributes):
            tmp = deepcopy(currentTransactionObject)
            # Find sender and recipient
            recipientIdx = None
            senderIdx = None
            amountIdx = None
            for idx,attr in enumerate(transferAttributes[start:]):
                if attr["key"] == "recipient":
                    recipientIdx = idx + start
                if attr["key"] == "sender":
                    senderIdx = idx + start
                if attr["key"] == "amount":
                    amountIdx = idx + start
                    # We need at least a sender or recipient and an amount
                    assert(recipientIdx!=None or senderIdx!=None)
                    assert(amountIdx!=None)
                    assert(amountIdx > 0)
                    # We stop here as we have found one set, amount should always be the last one
                    start = start + idx + 1
                    break
            
            # Fill to and from address
            if recipientIdx is not None:
                tmp.toAddress = transferAttributes[recipientIdx]["value"]
            if senderIdx is not None:
                tmp.fromAddress = transferAttributes[senderIdx]["value"]
            
            # Parse amounts
            amountPositions = self._parseAmounts(transferAttributes[amountIdx]["value"])
            for amount in amountPositions:
                tmp2 = deepcopy(tmp)
                #check for in or out transaction
                if tmp2.fromAddress == self.address:
                    tmp2.posOut = amount
                    txList.append(tmp2)
                elif tmp2.toAddress == self.address:
                    tmp2.posIn = amount
                    txList.append(tmp2)
                
        # Finally declutter the list (find and remove opposite transactions)
        for txElem in txList:
            for txElem2 in txList:
                if (txElem.posIn == txElem2.posOut and txElem.posIn.amount != 0) or (txElem.posOut == txElem2.posIn and txElem.posOut.amount != 0): #and txElem.fromaddress == txElem2.toAddress:
                    #print("REMOVING in tx hash {} where posIn1 ={}, posOut2={}, posOut1={}, posIn2={}".format(txElem.txHash, txElem.posIn.amount, txElem2.posOut.amount, txElem.posOut.amount, txElem2.posIn.amount))
                    txList.remove(txElem)
                    txList.remove(txElem2)
        
        # Find trades
        for txElem in txList:
            for txElem2 in txList:
                if txElem.fromAddress == txElem2.toAddress and txElem.toAddress == txElem2.fromAddress:
                    txElem.combinePositions(txElem2)
                    txElem.category = "Swap"
                    txList.remove(txElem2)
        if len(txList) == 2 and (txList[0].posIn.amount == 0 or txList[0].posOut.amount == 0) and (txList[1].posIn.amount == 0 or txList[1].posOut.amount == 0):
            txList[0].combinePositions(txList[1])
            txElem.category = "Swap"
            txList.remove(txElem2)
            
        return txList


    def getTransactions(self, startTime=None, offset=None) -> TransactionList:
        self.rawTxs = self._getAllRawTransactions()

        for tx in self.rawTxs:
            self.chain = tx["header"]["chain_id"]
            time = datetime.strptime(tx["header"]["timestamp"],"%Y-%m-%dT%H:%M:%SZ")
            height = tx["data"]["height"]
            txHash = tx["data"]["txhash"]

            t = Transaction(dateTime=time, txHash=txHash, blockHeight=height)
            #print(txHash)
            #1. Fee
            try:
                fee = self._returnValueFromPath(tx, self.pathToFee)
                feeCurrency = self._returnValueFromPath(tx, self.pathToFeeCurrency)
                t.fee = Fee(Decimal(fee)/self.denominator, self.getSymbol(feeCurrency) )
            except IndexError:
                print("No fee found for transaction {}".format(txHash))

            #2.Memo
            t.memo = self._returnValueFromPath(tx, self.pathToMemo)

            #3. Check for failed transaction
            if not("logs" in tx["data"]):
                t.category = "Loss"
                if t.fee.amount > 0:
                    self.txList.append(t)
                continue 

            #4. Loop all logs and events
            txs = TransactionList()
            for log in tx["data"]["logs"]:
                tmp = deepcopy(t)
                events = log["events"] #todo: check --> no logs for failed txs
                txs += self.handleEvents(events, tmp)

            # Remove doubled fees, fee is only added per transaction and not per log or per event
            feeAdded = False
            for tx in txs:
                if tx.fee.amount > 0 and not(feeAdded):
                    feeAdded = True
                elif tx.fee.amount > 0 and feeAdded:
                    tx.fee.amount = 0

            self.txList.extend(txs)
        return self.txList


from pycoin.networks.registry import network_for_netcode
from pycoin.networks.bitcoinish import create_bitcoinish_network
import blockcypher

from pyCryptoTransactions.networks.bitcoin import BitcoinImport

class LitecoinImport(BitcoinImport):
    Network = network_for_netcode("LTC")

    def __init__(self, publicKey):

        # In case a xpub instead of a Ltub key is used
        if publicKey[0:4].lower() == "xpub":
            self.Network = create_bitcoinish_network(
                        network_name="Litecoin", symbol="LTC", subnet_name="mainnet",
                        wif_prefix_hex="b0", sec_prefix="LTCSEC:", address_prefix_hex="30", pay_to_script_prefix_hex="32",
                        bip32_prv_prefix_hex="019d9cfe", bip32_pub_prefix_hex="0488B21E", bech32_hrp="ltc") # changed: bip32_pub_prefix_hex from bitcoin
        
        super().__init__(publicKey)
    






from pycoin.symbols.btc import network as BTC
from pycoin.convention import satoshi_to_btc
import requests
import time
import blockcypher
from enum import Enum

from pyCryptoTransactions.networks.Network import Network
from pyCryptoTransactions.Transaction import Position, Transaction, TransactionList, Fee
from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Coin import Coin
class AddresType(Enum):
    LEGACY = 0
    SEGWIT = 1
    NATIVE_SEGWIT = 2
    

class BitcoinImport(Importer):
    Network = BTC

    def __init__(self, publicKey):
        '''
        The __init__ function is called when an instance of the class is created
        :param publicKey: The public key of the user
        '''
        super().__init__()
        self.publicKey = publicKey
        self._blockExplorerUrl = "http://blockchain.info"
        self.coin = Coin(self.Network.symbol, self.Network.network_name)
    
    def _getSubkey(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        '''
        It takes the public key and the desired subkey path as arguments and returns the subkey
        
        :param receiveAddr: True if you want to receive coins, False if you want to change coins, defaults
        to True
        :type receiveAddr: bool (optional)
        :param changeAddr: If True, return the change address for the account. If False, return the
        receiving address, defaults to False
        :type changeAddr: bool (optional)
        :param number: the index of the subkey to retrieve, defaults to 0 (optional)
        :return: The subkey object.
        '''
        if receiveAddr == changeAddr:
            raise ValueError("Either choose receiving address or change address, not both!")

        key = self.Network.parse(self.publicKey) #direct version (xpub=legacy, ypub=segwit, zpub=native segwit)
        subkey = key.subkey_for_path("{0}/{1}".format("0" if receiveAddr else "1", str(number)))
        return subkey

    def generateAddresses(self, func, receiveAddr:bool=True, changeAddr:bool=False, numAddresses=1000):
        '''
        Generate a list of addresses
        
        :param func: the function that generates the address
        :param receiveAddr: if True, generate a receive address. If False, then generate a change
        address, defaults to True
        :type receiveAddr: bool (optional)
        :param changeAddr: if True, generate a change address, defaults to False
        :type changeAddr: bool (optional)
        :param numAddresses: the number of addresses to generate, defaults to 1000 (optional)
        :return: A generator object.
        '''
        # returns a generator object
        counter = 0
        while True:
            if (counter > numAddresses): 
                return
            yield func(receiveAddr, changeAddr, counter)
            counter += 1

    def generateAddress(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        '''
        If you want to generate an address, call this function
        
        :param receiveAddr: If True, the address will be marked as a receiving address. If False, it will be
        marked as a change address, defaults to True
        :type receiveAddr: bool (optional)
        :param changeAddr: If True, the address will be for change (rather than receive), defaults to False
        :type changeAddr: bool (optional)
        :param number: The index of the subkey to generate, defaults to 0 (optional)
        :return: The address of the subkey.
        '''
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        return subkey.address()

    def generateSegwitAddressFromXpub(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        "For Ledger Xpubs that are actually used as ypubs. Alternative: Convert to ypub: https://jlopp.github.io/xpub-converter/"
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        #(#P2WPKH in P2SH / ypub) -->pycoin/key/BIP49Node.py
        script = self.Network.contract.for_p2pkh_wit(subkey.hash160(is_compressed=True))
        return self.Network.address.for_p2s(script)

    def generateNativeSegwitAddressFromXpub(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        "For Ledger Xpubs that are actually used as ypubs"
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        #(#P2WPKH) -->pycoin/key/BIP84Node.py
        return self.Network.address.for_p2pkh_wit(subkey.hash160(is_compressed=True))
    
    def getNumTxForAddress(self, address, txData):
        pass

    def _getRawTxsForAddress(self, address, offset=None):
        print("Querying address {}".format(address))
        return blockcypher.get_address_details(address, coin_symbol=self.Network.symbol.lower())
    
    def _getRawTxDetails(self, txHash):
        #print("Querying Tx {}".format(txHash))
        return blockcypher.get_transaction_details(txHash, coin_symbol=self.Network.symbol.lower())

    def _query(self, url):
        '''
        Query the data. Takes a url as an input, calls the url, and returns the data
        
        :param url: the url to query
        :return: A dictionary of data.
        '''
        response = requests.get(url)
        print("Calling {}".format(url))
        if response.status_code != 200:
            raise BaseException("Wrong output")
        print(response.ok)
        print(response.status_code)
        data = response.json()
        time.sleep(10.5) #rate limit blockchain.info
        return data
    
    def _getTransactions(self, addressGenerator, offset=None, startTime=None, receiving=True):
        while(True):
            address = next(addressGenerator)
            data = self._getRawTxsForAddress(address, offset)
            num = data["n_tx"]
            if num == 0:
                break
            for elem in data["txrefs"]: #txs blockchain info
                txHash = elem["tx_hash"]
                change = satoshi_to_btc(elem["value"])
                # Check for out or inbound transaction
                isOut = elem["tx_output_n"] < 0
                isIn = elem["tx_output_n"] >= 0
                txFound = False

                if self.txList.hasTx(txHash):
                    t, idx = self.txList.fromTxHash(txHash)
                    txFound = True
                    #self.txList[idx] = t
                else:
                    # Fill new transaction object
                    t = Transaction()
                    t.datetime = elem["confirmed"]
                    t.txHash = txHash
                    
                if isIn and receiving:
                        t.posIn += Position(change, self.coin)
                if isOut:
                        t.posOut += Position(change, self.coin)
                if isIn and not(receiving):
                        t.posOut += Position(-change, self.coin)
                
                if not(txFound):
                    self.txList.append(t)


        return self.txList

    def _getAllTransactions(self, addressType:AddresType = AddresType.LEGACY):

        if addressType == AddresType.LEGACY:
            func = self.generateAddress

        elif addressType == AddresType.SEGWIT:
            func = self.generateSegwitAddressFromXpub
        
        elif addressType == AddresType.NATIVE_SEGWIT:
            func = self.generateNativeSegwitAddressFromXpub

        addressGenerator = self.generateAddresses(func, receiveAddr=True, changeAddr=False) 
        self._getTransactions(addressGenerator)
        addressGenerator = self.generateAddresses(func, receiveAddr=False, changeAddr=True) 
        self._getTransactions(addressGenerator, receiving=False)
        
        self.txList.sort(key=lambda r: r.datetime)


    def getTransactions(self, startTime=None, offset=None) -> 'TransactionList':
        #We test all 3 address types in case we have a ypub/zpub key as xpub key (as the Ledger software is doing)
        self._getAllTransactions(addressType=AddresType.LEGACY)
        time.sleep(5)
        self._getAllTransactions(addressType=AddresType.SEGWIT)
        time.sleep(5)
        self._getAllTransactions(addressType=AddresType.NATIVE_SEGWIT)

        #Sort
        self.txList.sort(key=lambda r: r.datetime)

        # Add fees only to outgoing transactions
        for tx in self.txList:
            if tx.isOutBound() and tx.fee.amount == 0:
                details = self._getRawTxDetails(tx.txHash)
                tx.fee = Fee( satoshi_to_btc(details["fees"]), self.Network.symbol)
            print(tx)

        return self.txList
        



from etherscan.accounts import Account
from etherscan.client import EmptyResponse
import json
import datetime
from django.utils import timezone
#from ..portfolioConfig import BSC_EXPLORER_APIKEY
BSC_EXPLORER_APIKEY = "8NJAS8T7PCVQXKID9NW72AB6KX2XI8BKQE"


class AccountBsc(Account):
    PREFIX = "https://api.bscscan.com/api?"

class BscTransactions():
    def __init__(self,  bscAccountAddress, apiKey=BSC_EXPLORER_APIKEY):
        self.apiKey = apiKey
        self.address = bscAccountAddress.lower()
        self.bscScanAccountApi = AccountBsc(address=bscAccountAddress, api_key=self.apiKey)
        self.transactions = []
        self.erc20Transactions = []
        self.internalTranscations = []

    def update(self):
        try:
            self.transactions = self.bscScanAccountApi.get_transaction_page()
        except EmptyResponse:
            self.transactions = []
        try:
            self.erc20Transactions = self.bscScanAccountApi.get_transaction_page(erc20=True)
        except EmptyResponse:
            self.erc20Transactions = []
        try:
            self.internalTransactions = self.bscScanAccountApi.get_transaction_page(erc20=False, internal=True)
        except EmptyResponse:
            self.internalTransactions = []

    def getAllTransactions(self, startTime=None):
        self.update()
        normalTransactions = self.getTransactionsAsDict(erc20=False)
        erc20TransActions = self.getTransactionsAsDict(erc20=True)
        internalTransactions = self.getTransactionsAsDict(erc20=False, internal=True)
        allTransactions = normalTransactions + erc20TransActions + internalTransactions

        #print(allTransactions)

        # Check for swap:
        for normalTransaction in normalTransactions:
            for erc20TransAction in erc20TransActions:
                if normalTransaction["txId"] == erc20TransAction["txId"] and normalTransaction["time"] == erc20TransAction["time"]:
                    if erc20TransAction["category"] == "Trade":
                        raise("Unknown condition")
                    if erc20TransAction["category"] == "Transfer In":
                        erc20TransAction["category"] = "Swap"
                        erc20TransAction["fee"] = normalTransaction["fee"]
                        erc20TransAction["feeCurrency"] = normalTransaction["feeCurrency"]
                        erc20TransAction["amountOut"] = normalTransaction["amountOut"]
                        erc20TransAction["currencyOut"] = normalTransaction["currencyOut"]
                        allTransactions.remove(normalTransaction)
                    elif erc20TransAction["category"] == "Transfer Out":
                        print(erc20TransAction)
                        print("#####")
                        print(normalTransaction)
                        if "amountIn" in normalTransaction:
                        #if normalTransaction["txId"] == "0x7c1b0f2587e972cd54be7d840193b67f8bee6789b8cdba55607920d58f6ce5b1":
                        #    continue
                            erc20TransAction["category"] = "Swap"
                            erc20TransAction["amountIn"] = normalTransaction["amountIn"]
                            erc20TransAction["currencyOut"] = normalTransaction["currencyIn"]
                            allTransactions.remove(normalTransaction)
                        else:
                            # atomic swaps
                            # do not remove transaction, but add another one
                            erc20TransAction["category"] = "Swap"
                            t = erc20TransAction.copy()
                            t["category"] = "Swap"
                            t["amountIn"] = t["amountOut"]
                            t["currencyIn"] = t["currencyOut"]
                            t["fee"] = 0
                            t["amountOut"] = 0
                            t["note"] = "Atomic Swap"
                            allTransactions.append(t)

        # Remove elements that are before startTime
        # for transaction in allTransactions:

        return allTransactions

    def isTransactionInERC20Transactions(self, transaction):
        for erc20Transaction in self.erc20Transactions:
            if erc20Transaction["hash"] == transaction["hash"]:
                return True
        return False

    def getNumberOfInputAndOutputTokenTransactions(self, transaction):
        inputs = 0
        outputs = 0
        numTokens = 0
        tokenList = []
        for erc20Transaction in self.erc20Transactions:
            if erc20Transaction["hash"] == transaction["hash"]:
                if erc20Transaction["from"] == self.address and float(erc20Transaction["value"]) > 0:
                    outputs += 1
                if erc20Transaction["to"] == self.address and float(erc20Transaction["value"]) > 0:
                    inputs += 1
                if erc20Transaction["tokenSymbol"] not in tokenList:
                    numTokens += 1
                    tokenList.append(erc20Transaction["tokenSymbol"])
                    print(tokenList)
        return (inputs, outputs, numTokens)

    def isDoubleInputTokenTransaction(self, transaction):
        inputs, outputs, _  = self.getNumberOfInputAndOutputTokenTransactions(transaction)
        return inputs == 2 and outputs < 2

    def isDoubleOutputTokenTransaction(self, transaction):
        inputs, outputs, _ = self.getNumberOfInputAndOutputTokenTransactions(transaction)
        return outputs == 2 and inputs < 2

    def isAddedLiquidity(self, transaction):
        inputs, outputs, numTokens = self.getNumberOfInputAndOutputTokenTransactions(transaction)
        if inputs == 1 and outputs == 2 and numTokens == 3:
            return True
        return False

    def isRemovedLiquidity(self, transaction):
        inputs, outputs, numTokens = self.getNumberOfInputAndOutputTokenTransactions(transaction)
        if inputs == 2 and outputs == 1 and numTokens == 3:
            return True
        return False

    def isStakingCompound(self, transaction):
        # 2 Inputs and 1 Output (e.g. Cake - Cake Syrup: Cake (reward) in, Cake(reward) out, Sryup (in)
        inputs, outputs, numTokens = self.getNumberOfInputAndOutputTokenTransactions(transaction)
        if inputs == 2 and outputs == 1 and numTokens == 2:
            return True
        return False

    def getTransactions(self, erc20=False, internal=False):
        if erc20:
            return self.erc20Transactions
        elif internal:
            return self.internalTransactions
        else:
            return self.transactions

    def getTransactionsAsDict(self, erc20=False, internal=False):
        transactionList = []
        for transaction in self.getTransactions(erc20, internal):
            entry = {}
            value = int(transaction["value"])/1E18 if not erc20 else int(transaction["value"]) / (10**int(transaction["tokenDecimal"]))
            currency = "BNB" if not erc20 else transaction["tokenSymbol"]

            if transaction["from"] == self.address and value > 0:
                entry["category"] = "Transfer Out"
                entry["amountOut"] = value
                entry["currencyOut"] = currency
                # todo check for erc20 token with same txid, then its a swap! ###or: search in both output transactions fpr same id!!

            elif transaction["to"] == self.address and value > 0:
                entry["category"] = "Transfer In"
                entry["amountIn"] = value
                entry["currencyIn"] = currency
                # todo check for erc20 token with same txid, then its a swap!

            elif transaction["from"] == self.address and value == 0:
                # Check if an complimentary entry exists in the ERC20 transactions
                # Then skip, because fees are contained in that
                if not(erc20) and self.isTransactionInERC20Transactions(transaction):
                    continue
                entry["category"] = "Fees"
            elif transaction["to"] == self.address and value == 0:
                continue
            else:
                print(transaction)
                print(value)
                print(transaction["to"] == self.address)
                raise ("Unknown transaction!")

            entry["txId"] = transaction["hash"]
            entry["time"] = datetime.datetime.fromtimestamp(int(transaction["timeStamp"]), tz=timezone.utc)

            if not(erc20) and bool(int(transaction["isError"])):
                entry["category"] = "Loss"

            if erc20 and len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"] and \
                    transactionList[-1]["category"] != entry["category"] and not(self.isAddedLiquidity(transaction)) \
                    and not(self.isRemovedLiquidity(transaction)):
                # Check that the currencies match (must be different in a trade)
                if (("amountIn" in entry and entry["amountIn"] > 0) and \
                    ("amountOut" in transactionList[-1] and transactionList[-1]["amountOut"] > 0) and
                    (entry["currencyIn"]!=transactionList[-1]["currencyOut"])) or \
                    (("amountOut" in entry and entry["amountOut"] > 0) and \
                     ("amountIn" in transactionList[-1] and transactionList[-1]["amountIn"] > 0) and
                     (entry["currencyOut"] != transactionList[-1]["currencyIn"])):

                    lastEntry = transactionList.pop()
                    entry = {**lastEntry, **entry} # combines both dict entries
                    entry["category"] = "Trade"

                # TODO: double transactions can also be compound + harvest ---> they are wrong now (trade + transfer in)

            ## Add liquidity = two output transactions
            #if erc20 and len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"] and \
            #        transactionList[-1]["category"] == entry["category"] and entry["category"] == "Transfer Out":
            ##    # add another transaction
            #    print(entry["txId"])
            #    #entry["category"] = "Liquidity"
            #    #transactionList[-1]["category"] = "Liquidity"
            #    print("LIQUIDITY added")

            ## Remove liquidity
            #if erc20 and len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"] and \
            #        transactionList[-1]["category"] == entry["category"] and entry["category"] == "Transfer in":
            ##    # add another transaction
            #    print(entry["txId"])
            #    #entry["category"] = "Liquidity"
            #    #transactionList[-1]["category"] = "Liquidity"
            #    print("LIQUIDITY removed")

            if entry["category"] != "Transfer In":
                entry["fee"] = int(transaction["gasUsed"]) * int(transaction["gasPrice"]) / 1E18
                entry["feeCurrency"] = "BNB"

            if erc20 and self.isAddedLiquidity(transaction):
                entry["category"] = "Added Liquidtiy"
                if len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"]:
                    entry["fee"] = 0

            if erc20 and self.isRemovedLiquidity(transaction):
                entry["category"] = "Removed Liquidity"
                if len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"]:
                    entry["fee"] = 0

            #if erc20 and self.isStakingCompound(transaction):
            #    entry["category"] = "Compound"
            #    if len(transactionList) > 0 and transactionList[-1]["txId"] == entry["txId"]:
            #        entry["fee"] = 0

            #if erc20 and "currencyOut" in entry and entry["currencyOut"].upper() == "CAKE-LP" and entry["category"]=="Tramsfer Out":
            #    entry["category"] = "Farm add"

            #if erc20 and entry["currencyIn"].upper() == "CAKE-LP":
            #    entry["category"] = "Farm remove"

            if internal:
                entry["note"] = "Internal Tx"

            transactionList.append(entry)

        return transactionList
    
    def getBalances(self):
        transactions = self.getAllTransactions()

        print("FINISHED FETCHING")

        print(transactions)

        assetList = []
        for transaction in transactions:
            if "currencyIn" in transaction and not(transaction["currencyIn"] in assetList) and transaction["currencyIn"] != "":
                assetList.append(transaction["currencyIn"])

        amountList = []
        for asset in assetList:
            elem = {}
            elem['currency'] = asset
            elem['amount'] = 0.0
            for transaction in transactions:
                if "currencyIn" in transaction and transaction["currencyIn"] == asset:
                    elem['amount'] += transaction["amountIn"]
                if "currencyOut" in transaction and transaction["currencyOut"] == asset:
                    elem['amount'] -= transaction["amountOut"]
                if "feeCurrency" in transaction and transaction["feeCurrency"] == asset:
                    elem['amount'] -= transaction["fee"]
            amountList.append(elem)

        return amountList

    # TODO: multiple inupts:  (beefy barn for example)
    # TODO: multiple outputs: wrong fee (doubled fee)

if __name__ == "__main__":
    print("main")
    testAcc = BscTransactions("0x28D736A96649F1ad228865bc2a0FF10Ec663a0b0")
    for asset in testAcc.getBalances():
        print("%s : %.2f\n" % (asset["currency"],asset["amount"]) )
    


#from ..models import Transaction
from django.utils import timezone
import datetime
from requests import Request, Session
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import json
import urllib3

#local
#iotaAddressesWithoutChecksum[i] = '"' .. string.sub(v, 0, 81) .. '"'
#postContent = '{"command":"getBalances","addresses":['..strjoin(',', iotaAddressesWithoutChecksum)..
#'],"threshold":100}'
#print(postContent)
#content = connection:request("POST", iotaRequestUrl(), postContent, "application/json", headers)

#function iotaRequestUrl()
#    return "https://nodes.thetangle.org:443"

MYADDRESSES = ['FTPEQZHWPO9BPPBDKNFFLFRORFVBDSJMSEPD9LESLTXJY9WHWPHCTFJHYLWNNCOWRXZXCXZEDKTJXKCAABXLAPPRPZ']#,\
              #'JBFNSFJAWDKNXYZOVMKZDKQIPNYQYIM9RQKCBHQRHAAYYAAZFT9X9YJ9WUCEVMQXXMIDFSCGYSVALSTNZDTXUTIBNX',\
            #'9VTPZAWYEPU9SHTYBOWCWZBWDIBBKMOMDOESIXYICJVBU9SRKZMMJMLRRRFZSFSSEXQDIWIHOKZTKXKYXCRCBD9HVA',\
            #'AUJLDOLKALNCMWVKCCKIGMLXEFXKQNGZGVNMFDOL9FTKXWDJRSDCLTNZWPJGXYDBKVGNB9NTTQFFECBF9TBWKUNNMC']

class IOTAImporter(object):

    def __init__(self):
        self._apiAdress = 'https://nodes.thetangle.org:443'
        self._initRequest()
        self.addAddresses = MYADDRESSES

    def _initRequest(self):
        headers = {'Accepts': 'application/json',  'X-IOTA-API-Version': '1'}
        self.session = Session()
        self.session.headers.update(headers)

    def getTransactions(self):
        for address in self.addAddresses:
            iotaAddressWithoutCheckSum = address[0:81]
            command = {
                "command": "getBalances",
                "addresses": [iotaAddressWithoutCheckSum]
            }
            stringified = json.dumps(command)
            print(stringified)
            data = self._request(path='',**command)
            print(data)

    def test(self):
        command = {
            "command": "getBalances",
            "addresses": [
                "FTPEQZHWPO9BPPBDKNFFLFRORFVBDSJMSEPD9LESLTXJY9WHWPHCTFJHYLWNNCOWRXZXCXZEDKTJXKCAA"
            ]
        }

        command = json.dumps(command)

        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        http = urllib3.PoolManager()

        response = http.request(
            'POST',
            'https://nodes.thetangle.org:443',
            headers={
                'Content-Type': 'application/json',
                'X-IOTA-API-Version': '1'
            },
            body=command)

        results = json.loads(response.data.decode('utf-8'))
        print(json.dumps(results, indent=1, sort_keys=True))

    def addAddress(self, address):
        #{"command": "getBalances", "addresses": ['..strjoin(', ', iotaAddressesWithoutChecksum)..
        # '],"threshold":100}'
        pass


    def _request(self, path, **params):
        try:
            print(params)
            print(self.session.headers)
            params = json.dumps(params)
            response = self.session.post(self._apiAdress, params=params)
            #print(response.url)
            data = json.loads(response.text)
            if data: #data['ok'] and data["code"]==0:
                return data
            else:
                raise("Data failure:")
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise("Connection Error")

iotaTest = IOTAImporter()
iotaTest.test()


from pandas.core.frame import DataFrame
from pyCryptoTransactions.Transaction import TransactionList
from pyCryptoTransactions.exporters.default import ExporterDefault
import datetime

#TODO: NOT VERY PYTHONIC, NEEDS IMPROVEMENT

class KoinlyExporter(ExporterDefault):

    def __init__(self, txData: TransactionList) -> None:
        super().__init__(txData)
        self.parseTxs()

    def parseTxs(self):
        data = []
        for tx in self.txData:
            
            #Only fee
            if tx.posIn.amount == 0 and tx.posOut.amount == 0 and tx.fee.amount != 0:
                entry = {"Date":datetime.datetime.strftime(tx.datetime, "%Y-%m-%d %H:%M"),
                        "Sent Amount":tx.fee.amount,
                        "Sent Currency":tx.fee.currency,
                        "Received Amount": "",
                        "Received Currency": "",
                        "Fee Amount":"",  #only show fee for trades
                        "Fee Currency":"",
                        "Net Worth Amount": "",
                        "Net Worth Currency": "",
                        "Label": "fee cost",
                        "Description": tx.note,
                        "TxHash":tx.txHash}
                data.append(entry)

            # In transaction with fee (e.g. Serum Dex)
            elif tx.posIn.amount != 0 and tx.posOut.amount == 0 and tx.fee.amount != 0:
                #create two entries, one for the fee as a cost
                entry = {"Date":datetime.datetime.strftime(tx.datetime, "%Y-%m-%d %H:%M"),
                        "Sent Amount":tx.fee.amount,
                        "Sent Currency":tx.fee.currency,
                        "Received Amount":"",
                        "Received Currency":"",
                        "Fee Amount":"",  #only show fee for trades
                        "Fee Currency":"",
                        "Net Worth Amount": "",
                        "Net Worth Currency": "",
                        "Label": "fee cost",
                        "Description": tx.note,
                        "TxHash":tx.txHash}
                entry2 = entry.copy()
                entry2["Received Amount"] = tx.posIn.amount
                entry2["Received Currency"] = tx.posIn.currency
                entry2["Sent Amount"] = ""
                entry2["Sent Currency"] = ""
                entry2["Label"] = tx.category
                data.append(entry)
                data.append(entry2)
            
            # Out transaction with a fee
            elif tx.posIn.amount == 0 and tx.posOut.amount != 0 and tx.fee.amount != 0:
                entry = {"Date":datetime.datetime.strftime(tx.datetime, "%Y-%m-%d %H:%M"),
                        "Sent Amount":tx.fee.amount,
                        "Sent Currency":tx.fee.currency,
                        "Received Amount":"",
                        "Received Currency":"",
                        "Fee Amount":"",  #only show fee for trades
                        "Fee Currency":"",
                        "Net Worth Amount": "",
                        "Net Worth Currency": "",
                        "Label": "fee  cost",
                        "Description": tx.note,
                        "TxHash":tx.txHash}
                entry2 = entry.copy()
                entry2["Sent Amount"] = tx.posOut.amount
                entry2["Sent Currency"] = tx.posOut.currency
                entry2["Label"] = tx.category
                data.append(entry)
                data.append(entry2)

            # In transaction without fee or out transaction without fee OR TRADE with fee
            #elif tx.posIn.amount != 0 or tx.posOut.amount !=0 and tx.fee.amount == 0 and (tx.posIn.amount == 0 or tx.posOut.amount == 0):
            else:
                entry = {"Date":datetime.datetime.strftime(tx.datetime, "%Y-%m-%d %H:%M"),
                        "Sent Amount":tx.posOut.amount if tx.posOut.amount!=0 else "",
                        "Sent Currency":tx.posOut.currency if tx.posOut.amount!=0 else "",
                        "Received Amount":tx.posIn.amount if tx.posIn.amount!=0  else "",
                        "Received Currency":tx.posIn.currency if tx.posIn.amount!=0 else "",
                        "Fee Amount":tx.fee.amount if tx.posIn.amount!=0 and tx.posOut.amount!=0 and tx.fee.amount!= 0 else "",  #only show fee for trades
                        "Fee Currency":tx.fee.currency if tx.posIn.amount!=0 and tx.posOut.amount!=0 and tx.fee.amount!=0 else "",
                        "Net Worth Amount": "",
                        "Net Worth Currency": "",
                        "Label": tx.category,
                        "Description": tx.note,
                        "TxHash":tx.txHash
                        }
                data.append(entry)
        self.df = DataFrame(data)




from .CoinbaseAuthentication import CoinbaseAuthentication
class CoinbaseAPIKeyAuthentication(CoinbaseAuthentication):
    '''
    Implements the authentication mechanism that uses the API key and secret.
    '''

    def __init__(self, key, secret):
        self.api_key = key
        self.api_secret = secret


    def get_data(self):
        return {
            'api_key': self.api_key,
            'api_secret': self.api_secret
        }



from requests import Request, Session
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import json
from datetime import datetime
import time
#from django.utils import timezone
import pytz as timezone
from decimal import Decimal
from copy import deepcopy

from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Coin import Coin

class BinanceChain(Importer):
    def __init__(self, address):
        super().__init__()
        self._apiAdress = 'https://dex.binance.org/api/v1'
        self._initRequest()
        self.rawTxs = None
        self.address = address
        self.denominator = Decimal(int(1E8))
        self.cancelledOrders = list()
    
    def _initRequest(self):
        headers = {'Accepts': 'application/json'}
        self.session = Session()
        self.session.headers.update(headers)
    
    def _request(self, path, **params):
        try:
            #print(params)
            response = self.session.get(self._apiAdress + path, params=params)
            #print(response.url)
            data = json.loads(response.text)
            if "ok" in data:
                if not(data["ok"]):
                    print(data)
                    raise("Error in the Request")
            if "code" in data:
                if data["code"] != 0:
                    print(data)
                    raise("Error in the request")
            if data: #data['ok'] and data["code"]==0:
                return data
            else:
                raise("Data failure:" + data['log'])
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise("Connection Error")
    
    # Endpoint doesn't give multisend transactions!
    def _getRawTransactions(self, offset=0, limit=500, **kwargs):#offset=0, limit=500, startTime=None, endTime = int(round(time.time() * 1000))):
        values = {}
        values['address'] = self.address
        values['offset'] = offset
        values['limit'] = limit
        for k in kwargs:
            if kwargs[k] is not None:
                values[k] = kwargs[k]
        self._request('/transactions', **values)

    def getTransaction(self, txId):
        """ Get a transaction for a given tx ID """
        #https://docs.binance.org/api-reference/dex-api/paths.html#apiv1transactions
        args = {}
        args['format'] = 'json'
        return self._request('/tx/' + txId, **args)
    
    def getOrder(self, orderId):
        args = {}
        args["format"] = 'json'
        return self._request('/orders/' + orderId, **args)
    
    def getTrades(self, orderId, timestamp:int, isBuy:bool):
        args = {}
        args["format"] = 'json'
        args["start"] = timestamp -10
        args["end"] = timestamp + 10
        if isBuy:
            args["buyerOrderId"] = orderId
        else:
            args["sellerOrderId"] = orderId
        print(args)
        return self._request('/trades', **args)

    def getFees(self):
        return self._request('/fees')

    #def getTrades(self,offset=0, limit=500, **kwargs):
    #    values = {}
    #    values['offset'] = offset
    #    values['limit'] = limit
    #    for k in kwargs:
    #        if kwargs[k] is not None:
    #            values[k] = kwargs[k]
    #    self._request('/trades', **values)


    def getTransactionsPage(self,page=-1):
        """get Raw Tx Data for the given address from the binance explorer. 
           Note: page must be negative!
        """
        # Get a page also inlcuding multisend transactions
        #https: // github.com / binance - chain / java - sdk / issues / 30
        #https://explorer.binance.org/api/v1/txs?page=-1&rows=100&address=ADDRESS
        params = {'address':self.address, 'page':page, 'rows':100}
        tmpApiAddress = self._apiAdress
        self._apiAdress = "https://explorer.binance.org/api/v1"
        ret = self._request('/txs', **params)
        self._apiAdress = tmpApiAddress
        return ret

    def _getAllRawTxsFromTxPage(self):
        # Query first page
        page = -1
        dataTransactionPage = self.getTransactionsPage(page=page)
        numTxs = dataTransactionPage["txNums"]
        numTxsQueried = 100
        while numTxsQueried < numTxs:
            page -= 1
            tmp = self.getTransactionsPage(page=page)
            numTxsQueried += 100
            print(type(tmp))
            dataTransactionPage.update(tmp)
            #d.update(tmp)
            #dataTransactionPage += dict(dataTransactionPage.items() | tmp.items())
        self.rawTxs = dataTransactionPage
        #todo: remove all txs over a given minimum time

    # currently unused
    def _getAllMultiSendTransactionIds(self):
        multiSendTxIds = []
        for entry in self.rawTxs:
            if entry["hasChildren"] > 0:
                multiSendTxIds.append(entry["txHash"])

        return multiSendTxIds

    def _getAllTransactionIds(self):
        ids = []
        for entry in self.rawTxs["txArray"]:
            ids.append((entry["timeStamp"],entry["txHash"],entry["txFee"]))
        return ids
    
    #def manualImportTransactionFromID(self, txID):
    #    tx = self.getTransaction(txID)
    #    t = Transaction()
    #    t.txHash = result["hash"]
    #    t.datetime = datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)

    def getTransactions(self, startTime=None, offset=None) -> 'TransactionList':
        self._getAllRawTxsFromTxPage()
        for timestamp, id, feeValue in self._getAllTransactionIds():
            feeValue = Decimal(str(feeValue))
            result = self.getTransaction(id)

            t = Transaction(network="binanceChain")

            #only transactions of type "auth/StdTx" are supported
            if result["tx"]["type"] != "auth/StdTx":
                print("Unsupported transaction with id {}".format(id))
                continue

            t.txHash = result["hash"]
            t.datetime = datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)

            #print(result["tx"]["value"]["msg"])
            # Only take first one, typically there are no more , TODO: Check or do better :)
            if result["tx"]["value"]["msg"][0]["type"] == "cosmos-sdk/Send":
                transaction = result["tx"]["value"]["msg"][0]["value"]

                isMulitSend = self.isMultiSendTransaction(result)
                isOutTransaction = self.isOutTransaction(result)

                if isOutTransaction:
                    t.category = "Transfer Out" if not(isMulitSend) else "Stake"
                    if feeValue>0:
                        t.fee = Fee(feeValue,"BNB")
                    else:
                        t.fee = Fee(Decimal(37500)/self.denominator if not(isMulitSend) else Decimal(30000)/self.denominator, "BNB")
                    coins = transaction["outputs"][0]["coins"]
                    #fixedFees = self.getFees()["fixed_fee_params"]  # Todo query fixed fees
                else:
                    t.category = "Transfer In" if not (isMulitSend) else "Harvest"
                    coins = transaction["inputs"][0]["coins"]

                if not(isMulitSend) and result["tx"]["value"]["memo"].find("SWAP") != -1 or result["tx"]["value"]["memo"].find("OUTBOUND") != -1:
                    t.category = "Swap"

                if result["tx"]["value"]["memo"]:
                    t.note = "memo:" + result["tx"]["value"]["memo"]

                i = 0
                for coin in coins:
                    tmp = deepcopy(t)
                    if isOutTransaction:
                        tmp.posOut = Position(Decimal(coin["amount"])/self.denominator, coin["denom"].split("-")[0])
                    else:
                        tmp.posIn = Position(Decimal(coin["amount"])/self.denominator, coin["denom"].split("-")[0]) 
                    if i>0:
                        tmp.fee.amount = Decimal(0)
                    self.txList.append(tmp)
                    i+=1
                    # Hack: Put equivalent transaction for BEPSWAP
                    #if tmp["category"] == "Stake":
                    #    tmp2 = tmp.copy()
                    #    tmp2["amountIn"] = tmp2["amountOut"]
                    #    tmp2["currencyIn"] = tmp2["currencyOut"]
                    #    tmp2["amountOut"] = 0
                    #    tmp2["category"] = "LP"
                    #    tmp2["wallet"] = "Bepswap"
                    #    transactios.append(tmp2)

            elif result["tx"]["value"]["msg"][0]["type"] == "dex/NewOrder": #BUY
                t.note = "dex/New Order"
                t.category = "Trade"
                order = result["tx"]["value"]["msg"][0]["value"]
                t.orderId = order["id"]
                orderInfo = self.getOrder(t.orderId)
                pair = order["symbol"].split('_')
                coin1 = Coin(pair[0].split('-'), contractAddress=pair[0])
                coin2 = Coin(pair[1].split('-'), contractAddress=pair[1])

                trades = self.getTrades(timestamp=int(t.datetime.timestamp()*1E3), orderId=t.orderId, isBuy=orderInfo["side"]==1)
                if order["ordertype"] == 2 and order["side"]==2: #sell
                    posIn = Position(Decimal(0),coin2)
                    posOut = Position(Decimal(0),coin1)
                    fee = Fee(0, "BNB")
                    for trade in trades["trade"]:
                        posIn += Position( Decimal(trade["price"].rstrip("0")) * Decimal(trade["quantity"].rstrip("0")), coin2)
                        posOut += Position( Decimal(trade["quantity"].rstrip("0")), coin1)
                    fee = Fee( Decimal(trade["sellFee"].split(";")[0].split("BNB:")[1]), "BNB")
                elif order["ordertype"] == 2 and order["side"]==1: #buy
                    posIn = Position(Decimal(0),coin1)
                    posOut = Position(Decimal(0),coin2)
                    for trade in trades["trade"]:
                        posIn += Position(  Decimal(trade["quantity"].rstrip("0")), coin1)
                        posOut += Position(Decimal(trade["price"].rstrip("0")) * Decimal(trade["quantity"].rstrip("0")), coin2)
                    fee = Fee( Decimal(trade["buyFee"].split(";")[0].split("BNB:")[1]), "BNB")

                t.posIn = posIn
                t.posOut = posOut
                t.fee = fee
                    
                #orderPrice1 = Decimal(order["price"])/self.denominator
                #orderQuantity1 = Decimal(order["quantity"])/self.denominator

               #print("{} == P1:{} Q1:{}, P2:{}, Q2:{} ({})".format(t.orderId, orderPrice1,orderQuantity1,orderPrice,orderQuantity,pair))

                if t.orderId in self.cancelledOrders or orderInfo["status"]=="Canceled":
                    continue
                #if order["ordertype"] == 2 and order["side"]==2: #sell
                #    t.posIn = Position(orderPrice * orderQuantity, pair[1].split('-')[0])
                #    t.posOut = Position(orderQuantity, pair[0].split('-')[0])
                #elif order["ordertype"] == 2 and order["side"]==1: #buy
                #    t.posIn = Position(orderQuantity, pair[0].split('-')[0])
                #    t.posOut = Position(orderPrice * orderQuantity, pair[1].split('-')[0])
                
                #t.fee = Fee(0, "BNB")
                t.price = Decimal(order["price"])/self.denominator
                t.tradingPair = (coin1, coin2)
                
                self.txList.append(t)

            elif result["tx"]["value"]["msg"][0]["type"] == "bridge/TransferOutMsg":
                # Swap to bsc network chain
                t.note = "Atomic Swap to BSC Chain"
                t.category = "Swap"
                t.posOut = Position(Decimal(result["tx"]["value"]["msg"][0]["value"]["amount"]["amount"]) / self.denominator, result["tx"]["value"]["msg"][0]["value"]["amount"]["denom"].split('-')[0])
                t.fee = Fee(Decimal("0.0002"), "BNB")
                self.txList.append(t)
            elif result["tx"]["value"]["msg"][0]["type"] == "dex/CancelOrder":
                # we must find orders that match this order id or prevent new ones getting added
                order = result["tx"]["value"]["msg"][0]["value"]
                orderId = order["refid"]
                if self.txList.fromOrderId(orderId) is not None:
                    self.txList.remove(self.txList.fromOrderId(orderId))
                else:
                    self.cancelledOrders.append(orderId)
                t.category = "Loss"
                t.fee = Fee(Decimal("0.00001"),"BNB")
                self.txList.append(t)
            else:
                raise("Unknown type: %s" % result["tx"]["value"]["msg"][0]["type"])

        return self.txList

    def isMultiSendTransaction(self, txResult):
        value = txResult["tx"]["value"]["msg"][0]["value"]
        if "inputs" in value and len(value["inputs"][0]["coins"]) > 1:
            return True
        if "outputs" in "value" and len(value["outputs"][0]["coins"]) > 1:
            return True
        return False

    def isOutTransaction(self, txResult):
        value = txResult["tx"]["value"]["msg"][0]["value"]
        if "inputs" in value and value["inputs"][0]["address"] == self.address:
            return True
        elif "outputs" in value and value["outputs"][0]["address"] == self.address:
            return False
        else:
            raise("Error: Input address not found in transacion entry!")

    #def get(self, address):
    #    # try to grab it all
    #    trades = self.getTrades()
    #    transactions = self._getRawTransactions(address=address, startTime=1609894800000, endTime=int(round(time.time() * 1000)))
    #    multiSendTransactions = []
    #    multiSendTransactionIds = self._getAllMultiSendTransactionIds(address=address)
    #    for id in multiSendTransactions:
    #        multiSendTransactions.append(self.getTransaction(id))

        # Now put it all together:
        # tbd








from pycoin.symbols.btc import network
from pycoin.networks.registry import network_for_netcode
import pycoin

key = network.parse.bip32("xprv9s21ZrQH143K31AgNK5pyVvW23gHnkBq2wh5aEk6g1s496M"
      "8ZMjxncCKZKgb5jZoY5eSJMJ2Vbyvi2hbmQnCuHBujZ2WXGTux1X2k9Krdtq")
print(key.hwif(as_private=1))
print(key.hwif())
print(key.wif())
print(key.sec())
print(key.address())

key = network.parse.bip32("xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63")
print(key.address())

key = network.parse.bip32_pub("xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63")
print(key.address()) #legacy
print(key.hash160()) #legacy
print(key.ku_output_for_address())

xpub = network.parse.address("xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP")
print("11")
print(xpub)

#legacy xpub (blockchain info)
key = network.parse("xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63")
print(key.address())
print(key.address(is_compressed=False))
subkey = key.subkey_for_path("0")
print(subkey.address())
subkey = key.subkey_for_path("0/0") #---->richtig
print(subkey.address())
subkey = key.subkey_for_path("1/0") #---->richtig
print(subkey.address())

key = network.parse("xpub661MyMwAqRbcGcWc7LfVdJnRUR2M9Qtp5675dw78aFLgjULMnniKvU4PhBve1gMHuVdUDZiykxdb2xxcqbxkZKNDssKwPZNJrJeP4j9z5fv")
subkey = key.subkey_for_path("0/0") 
print(subkey.address())

print("segwit")
#segwit
key = network.parse("xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP")
#nachfolgender geht, benoetigt aber conversion https://jlopp.github.io/xpub-converter/
#key = network.parse("ypub6WsFvFc3Wsp5EERuug41Pe5su4dGcVp9gsgtUGaqtTooRVfkfkcFmtahwsL3jCEUYQcCKDYv9NUyUVJDKxrfQhmfT2S7jpmtdG139Upvgub")
print(key.address())
print(key.address(is_compressed=False))
subkey = key.subkey_for_path("0")
print(subkey.address())
subkey = key.subkey_for_path("0/0") 
print(subkey.address())
subkey = key.subkey_for_path("1/0") 
print(subkey.address())
#39Abxbt4vs78qfiUw8wDd1cGr3Rg7wURaK

subkey = key.subkey_for_path("0/0") 
print("Native segwit")
print(network.address.for_p2pkh_wit(subkey.hash160(is_compressed=True)))
#Segwit:
print("Segwit")
print(network.address.for_p2sh(subkey.hash160(is_compressed=True)))

## From contract
script = network.contract.for_p2pkh_wit(subkey.hash160(is_compressed=True))
print(network.address.for_p2s(script)) #--->RICHTIG (#P2WPKH in P2SH / ypub)
#-->pycoin/key/BIP49Node.py

#Geht nicht:
#key = network.parse.bip49("xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP")
#subkey = key.subkey_for_path("0/0") 
#print(subkey.address())

#pycoin.networks._bip49_pub_prefix = "xpub"
#key = network.parse.bip49_pub("xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP")
#print(key.subkey_for_path("1/0").address())

##WORKS:
#~/C/crypto-/networks/bitcoin ❯ ku xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63 -s 0/0-10 -a
#ku xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63 -s 1/0-10 -a


## Litecoin from xpub
print("assdasd\n")
from pycoin.networks.bitcoinish import create_bitcoinish_network
network = create_bitcoinish_network(
    network_name="Litecoin", symbol="LTC", subnet_name="mainnet",
    wif_prefix_hex="b0", sec_prefix="LTCSEC:", address_prefix_hex="30", pay_to_script_prefix_hex="32",
    bip32_prv_prefix_hex="019d9cfe", bip32_pub_prefix_hex="0488B21E", bech32_hrp="ltc") ##bip32_pub_prefix_hex from bitcoin
pub = "xpub661MyMwAqRbcFrfZqAH8d94hEQKBitWTBMNmshSZiVrydCE3dFRt8C6TwnufWTZjFGgEfoDPJTVGdbFSJFYYpZDRonDrtpGPyMkcGXhpmFX"
key = network.parse(pub)
print(key.subkey_for_path("0/0").address()) #-->stimmt


#litecoin
print("\nLitecoin")
network = network_for_netcode("LTC")
pub = "Ltub2Z1YoRfU7gavCJ5ChwGNAnmcpDK35wzCRAB2mstyVjV7ShQGnizKTyL6kFNhK4KNuWypAYvAS534LFowXrFRPLQRacRBemKrrAWCzuPEe4o"
key = network.parse.bip32(pub)
print(key.subkey_for_path("0/0").address()) #-->stimmt
#LbNrLRMKNG68zHeNnDdMotK8Vr2jLz3bej

#segwit
pub = "Ltub2ZYYmNuYZRAyCFNp8LAPEwxnRG9jkx4qmmPDwTfQNe19JgURHxAL1WjoMP24ZP8fxZafCTd8SBR9acr9KWjWmM3rtyFsRfQf3dk3cK2hgV5"
key = network.parse(pub)
subkey = key.subkey_for_path("0/0") 
script = network.contract.for_p2pkh_wit(subkey.hash160(is_compressed=True)) #xpub as ypub
print(network.address.for_p2s(script)) #--->stimmt
#MULiubAyQSZNDgvvtj7q5D64vMbCCJfafP


#########TESTS
#mobj = pycoin.mnemonic.Mnemonic("english")
#seed = mobj.to_seed("bench auction print immense castle slide tip pill evolve abandon canvas ugly")
#key = network.keys.bip39_seed("99b269aecb097d51961c086e2c6baa595d9e939e156709871352c737b91f55806d3ec11c27251caeec650215e0a02da4ca4a5e4f7659290060dc37ef2a0a1bc3")
#subkey = key.subkey_for_path("44/2H/0H/0/0")
#print(subkey.address)

# Loafwallet 
pub = "Ltub2XuZ5AKXavgGqmSegPmhyZ5BMS4FJAZoCweSRCUDBTffni8DuNfJxu8BjqinjSqKtpE1ED8mCNgKt2Y3gQKaTEup5TFdfjv9NXYMxMEvXeg"
key = network.parse.bip32(pub)
print(key.subkey_for_path("0").address()) #--->das ist er
print(key.address())

pub = "Ltub2Vf5gKHAw7d1hUfLu3iaLMvzkND99PoE2tr2duBdDvRMyUGG7zdaURJjGZGtFRLnzNEL1oSijkTfPfk86M3FkWE5p2zbAoHrsQ7vEfeyoEk"
key = network.parse.bip32(pub)
print(key.subkey_for_path("0/0").address()) #----> YES. LOAFWALLET: Derivation Path m/0'/ on https://iancoleman.io/bip39/ (Tab bip32)


#Electrum LTC: xpub661MyMwAqRbcFrfZqAH8d94hEQKBitWTBMNmshSZiVrydCE3dFRt8C6TwnufWTZjFGgEfoDPJTVGdbFSJFYYpZDRonDrtpGPyMkcGXhpmFX
#pub = "xpub661MyMwAqRbcFrfZqAH8d94hEQKBitWTBMNmshSZiVrydCE3dFRt8C6TwnufWTZjFGgEfoDPJTVGdbFSJFYYpZDRonDrtpGPyMkcGXhpmFX"
#key = network.parse(pub)
#print(key.subkey_for_path("0/0").address()) 
#print(key.address())

print(network.symbol)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyCryptoTransactions',
 'pyCryptoTransactions.exchanges',
 'pyCryptoTransactions.exchanges.binance',
 'pyCryptoTransactions.exchanges.coinbase',
 'pyCryptoTransactions.exchanges.coinbase.coinbaseApiWrapper.coinbase',
 'pyCryptoTransactions.exporters',
 'pyCryptoTransactions.manual',
 'pyCryptoTransactions.networks',
 'pyCryptoTransactions.networks.binanceChain',
 'pyCryptoTransactions.networks.bitcoin',
 'pyCryptoTransactions.networks.bitcoin.Archive',
 'pyCryptoTransactions.networks.bsc',
 'pyCryptoTransactions.networks.cosmos',
 'pyCryptoTransactions.networks.eth',
 'pyCryptoTransactions.networks.iota',
 'pyCryptoTransactions.networks.litecoin',
 'pyCryptoTransactions.networks.osmos',
 'pyCryptoTransactions.networks.persistence',
 'pyCryptoTransactions.networks.thorchain']

package_data = \
{'': ['*'], 'pyCryptoTransactions.exchanges.coinbase': ['coinbaseApiWrapper/*']}

install_requires = \
['blockcypher>=1.0.93,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'pycoin>=0.91.20210515,<0.92.0',
 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'pycryptotransactions',
    'version': '0.1.3',
    'description': 'Python Package to query crpyto transactions from various blockchains and exchanges through public APIs',
    'long_description': '# py-Crypto-Transactions\n\n[![PyPI](https://badge.fury.io/py/pycryptotransactions.svg)](https://badge.fury.io/py/pycryptotransactions)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/pycryptotransactions)\n![GitHub](https://img.shields.io/github/license/pcko1/pycryptotransactions)\n[![Python 3.8](https://img.shields.io/badge/python-3.10-blue.svg)](https://www.python.org/downloads/release/python-310/)\n![GitHub Issues](https://img.shields.io/github/issues/jensb89/py-Crypto-Transactions)\n\n\nA python package to query crypto transaction from various blockchains and exchanges\nto build your own portfolio app!\n\npyCryptoTransactions also supports a few exporters, such that data can be imported e.g. at Koinly or Accointing.\n\nNote: Very elary alpha version!\n\n## Install\nUse pip:\n * pip install pyCryptoTransactions (NOTE: make sure to use camelCase name like this and not pycryptotransactions)\nOr use poetry \nOr install directly by cloning the git repo\n## Supported Blockchains\n * Bitcoin\n * Litecoin\n * Cosmos\n * Thorchain\n * Osmos\n * Iota\n * Persistence\n * Ethereum (soon)\n * Bsc (soon)\n\n## Supported exchanges\n * Binance (support for exported CSVs, no API yet)\n * Coinbase (Oauth Api)\n * Coinbase Pro (soon)\n\n## Examples\n```\nt = LitecoinImport("xpub...")\nt.getTransactions()`\n```\n\n```\nt = CosmosChain("cosmos....")\ntxs = t.getTransactions()\nprint(txs)\ndf = txs.toPandasDataframe()\ndf.to_csv("atom_test.csv")\nprint(txs.calculateBalance())\na = AccointingExporter(txs)\na.exportToExcel("atom_accointing.xlsx")\n```',
    'author': 'Jens Brauer',
    'author_email': 'jensb89@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jensb89/py-Crypto-Transactions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
